import os
from pathlib import Path
from typing import Collection

import numpy as np
import pandas as pd
import yaml
from larch import DataFrames, Model, P, X
from larch.util import Dict

from .general import (
    apply_coefficients,
    construct_nesting_tree,
    cv_to_ca,
    explicit_value_parameters,
    linear_utility_from_spec,
    remove_apostrophes,
    str_repr,
)


def schedule_choice_model(
    name,
    edb_directory="output/estimation_data_bundle/{name}/",
    coefficients_file="{name}_coefficients.csv",
    spec_file="{name}_SPEC.csv",
    alt_values_file="{name}_alternatives_combined.csv",
    chooser_file="{name}_choosers_combined.csv",
    settings_file="{name}_model_settings.yaml",
    return_data=False,
):
    model_selector = name.replace("_location", "")
    model_selector = model_selector.replace("_destination", "")
    model_selector = model_selector.replace("_subtour", "")
    model_selector = model_selector.replace("_tour", "")
    edb_directory = edb_directory.format(name=name)

    def _read_csv(filename, optional=False, **kwargs):
        filename = filename.format(name=name)
        try:
            return pd.read_csv(os.path.join(edb_directory, filename), **kwargs)
        except FileNotFoundError:
            if optional:
                return None
            else:
                raise

    settings_file = settings_file.format(name=name)
    with open(os.path.join(edb_directory, settings_file), "r") as yf:
        settings = yaml.load(
            yf,
            Loader=yaml.SafeLoader,
        )

    try:
        coefficients = _read_csv(
            coefficients_file,
            index_col="coefficient_name",
        )
    except FileNotFoundError:
        # possibly mis-named file is shown in settings
        coefficients_file = settings.get("COEFFICIENTS", coefficients_file)
        coefficients = _read_csv(
            coefficients_file,
            index_col="coefficient_name",
        )

    spec = _read_csv(spec_file, comment="#")
    alt_values = _read_csv(alt_values_file)
    chooser_data = _read_csv(chooser_file)

    # remove temp rows from spec, ASim uses them to calculate the other values written
    # to the EDB, but they are not actually part of the utility function themselves.
    spec = spec.loc[~spec.Expression.str.startswith("_")].copy()

    include_settings = settings.get("include_settings")
    if include_settings:
        with open(os.path.join(edb_directory, include_settings), "r") as yf:
            more_settings = yaml.load(
                yf,
                Loader=yaml.SafeLoader,
            )
        settings.update(more_settings)

    CHOOSER_SEGMENT_COLUMN_NAME = settings.get("CHOOSER_SEGMENT_COLUMN_NAME")
    SEGMENT_IDS = settings.get("SEGMENT_IDS")
    if SEGMENT_IDS is None:
        SEGMENTS = settings.get("SEGMENTS")
        if SEGMENTS is not None:
            SEGMENT_IDS = {i: i for i in SEGMENTS}

    if "Label" in spec.columns:
        label_column_name = "Label"
    elif "Expression" in spec.columns:
        label_column_name = "Expression"
    else:
        raise ValueError("cannot find Label or Expression in spec file")

    m = Model()
    if len(spec.columns) == 4 and (
        [c.lower() for c in spec.columns]
        == ["label", "description", "expression", "coefficient"]
    ):
        m.utility_ca = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=spec.columns[-1],
            ignore_x=("local_dist",),
        )
    elif (
        len(spec.columns) == 4
        and all(spec.columns[:3] == ["Label", "Description", "Expression"])
        and len(SEGMENT_IDS) == 1
        and spec.columns[3] == list(SEGMENT_IDS.values())[0]
    ):
        m.utility_ca = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=spec.columns[-1],
            ignore_x=("local_dist",),
        )
    else:
        m.utility_ca = linear_utility_from_spec(
            spec,
            x_col=label_column_name,
            p_col=SEGMENT_IDS,
            ignore_x=("local_dist",),
            segment_id=CHOOSER_SEGMENT_COLUMN_NAME,
        )

    apply_coefficients(coefficients, m, minimum=-25, maximum=25)

    chooser_index_name = chooser_data.columns[0]
    x_co = chooser_data.set_index(chooser_index_name)
    alt_values.fillna(0, inplace=True)
    x_ca = cv_to_ca(
        alt_values.set_index([chooser_index_name, alt_values.columns[1]]),
        required_labels=spec[label_column_name],
    )

    # if CHOOSER_SEGMENT_COLUMN_NAME is not None:
    #     # label segments with names
    #     SEGMENT_IDS_REVERSE = {v: k for k, v in SEGMENT_IDS.items()}
    #     x_co["_segment_label"] = x_co[CHOOSER_SEGMENT_COLUMN_NAME].apply(
    #         lambda x: SEGMENT_IDS_REVERSE[x]
    #     )
    # else:
    #     x_co["_segment_label"] = size_spec.index[0]

    alt_codes = np.arange(len(x_ca.index.levels[1])) + 1
    x_ca.index = x_ca.index.set_levels(alt_codes, 1)
    x_co["override_choice_plus1"] = x_co["override_choice"] + 1
    x_co["model_choice_plus1"] = x_co["model_choice"] + 1

    unavail_coefs = coefficients.query("(constrain == 'T') & (value < -900)").index
    unavail_data = [i.data for i in m.utility_ca if i.param in unavail_coefs]
    if len(unavail_data):
        joint_unavail = "|".join(f"({i}>0)" for i in unavail_data)
        joint_avail = f"~({joint_unavail})"
    else:
        joint_avail = 1

    d = DataFrames(co=x_co, ca=x_ca, av=joint_avail)
    m.dataservice = d
    m.choice_co_code = "override_choice_plus1"
    # m.choice_co_code = "model_choice_plus1"

    if return_data:
        return (
            m,
            Dict(
                edb_directory=Path(edb_directory),
                alt_values=alt_values,
                chooser_data=chooser_data,
                coefficients=coefficients,
                spec=spec,
                model_selector=model_selector,
                joint_avail=joint_avail,
            ),
        )

    return m


def construct_availability_ca(model, chooser_data, alt_codes_to_names):
    """
    Construct an availability dataframe based on -999 parameters.

    Parameters
    ----------
    model : larch.Model
    chooser_data : pandas.DataFrame
    alt_codes_to_names : Mapping[int,str]

    Returns
    -------
    pandas.DataFrame
    """
    avail = {}
    for acode, aname in alt_codes_to_names.items():
        unavail_cols = list(
            (
                chooser_data[i.data]
                if i.data in chooser_data
                else chooser_data.eval(i.data)
            )
            for i in model.utility_co[acode]
            if (i.param == "-999" or i.param == "-999.0")
        )
        if len(unavail_cols):
            avail[acode] = sum(unavail_cols) == 0
        else:
            avail[acode] = 1
    avail = pd.DataFrame(avail).astype(np.int8)
    avail.index = chooser_data.index
    return avail


def mandatory_tour_scheduling_work_model(return_data=False):
    return schedule_choice_model(
        name="mandatory_tour_scheduling_work",
        return_data=return_data,
        coefficients_file="tour_scheduling_work_coefficients.csv",
    )


def mandatory_tour_scheduling_school_model(return_data=False):
    return schedule_choice_model(
        name="mandatory_tour_scheduling_school",
        return_data=return_data,
        coefficients_file="tour_scheduling_school_coefficients.csv",
    )


def non_mandatory_tour_scheduling_model(return_data=False):
    return schedule_choice_model(
        name="non_mandatory_tour_scheduling",
        return_data=return_data,
    )


def joint_tour_scheduling_model(return_data=False):
    return schedule_choice_model(
        name="joint_tour_scheduling",
        return_data=return_data,
    )


def atwork_subtour_scheduling_model(return_data=False):
    return schedule_choice_model(
        name="atwork_subtour_scheduling",
        return_data=return_data,
    )



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.abm.models.trip_destination import run_trip_destination
from activitysim.abm.models.trip_purpose import run_trip_purpose
from activitysim.abm.models.util.trip import (
    cleanup_failed_trips,
    flag_failed_trip_leg_mates,
)
from activitysim.core import config, inject, pipeline, tracing
from activitysim.core.util import assign_in_place

from .util import estimation

logger = logging.getLogger(__name__)


def run_trip_purpose_and_destination(
    trips_df, tours_merged_df, chunk_size, trace_hh_id, trace_label
):

    assert not trips_df.empty

    choices = run_trip_purpose(
        trips_df,
        estimator=None,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=tracing.extend_trace_label(trace_label, "purpose"),
    )

    trips_df["purpose"] = choices

    trips_df, save_sample_df = run_trip_destination(
        trips_df,
        tours_merged_df,
        estimator=None,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=tracing.extend_trace_label(trace_label, "destination"),
    )

    return trips_df, save_sample_df


@inject.step()
def trip_purpose_and_destination(trips, tours_merged, chunk_size, trace_hh_id):

    trace_label = "trip_purpose_and_destination"
    model_settings = config.read_model_settings("trip_purpose_and_destination.yaml")

    # for consistency, read sample_table_name setting from trip_destination settings file
    trip_destination_model_settings = config.read_model_settings(
        "trip_destination.yaml"
    )
    sample_table_name = trip_destination_model_settings.get(
        "DEST_CHOICE_SAMPLE_TABLE_NAME"
    )
    want_sample_table = (
        config.setting("want_dest_choice_sample_tables")
        and sample_table_name is not None
    )

    MAX_ITERATIONS = model_settings.get("MAX_ITERATIONS", 5)

    trips_df = trips.to_frame()
    tours_merged_df = tours_merged.to_frame()

    if trips_df.empty:
        logger.info("%s - no trips. Nothing to do." % trace_label)
        return

    # FIXME could allow MAX_ITERATIONS=0 to allow for cleanup-only run
    # in which case, we would need to drop bad trips, WITHOUT failing bad_trip leg_mates
    assert MAX_ITERATIONS > 0

    # if trip_destination has been run before, keep only failed trips (and leg_mates) to retry
    if "destination" in trips_df:

        if "failed" not in trips_df.columns:
            # trip_destination model cleaned up any failed trips
            logger.info("%s - no failed column from prior model run." % trace_label)
            return

        elif not trips_df.failed.any():
            # 'failed' column but no failed trips from prior run of trip_destination
            logger.info("%s - no failed trips from prior model run." % trace_label)
            trips_df.drop(columns="failed", inplace=True)
            pipeline.replace_table("trips", trips_df)
            return

        else:
            logger.info("trip_destination has already been run. Rerunning failed trips")
            flag_failed_trip_leg_mates(trips_df, "failed")
            trips_df = trips_df[trips_df.failed]
            tours_merged_df = tours_merged_df[
                tours_merged_df.index.isin(trips_df.tour_id)
            ]
            logger.info("Rerunning %s failed trips and leg-mates" % trips_df.shape[0])

            # drop any previously saved samples of failed trips
            if want_sample_table and pipeline.is_table(sample_table_name):
                logger.info("Dropping any previously saved samples of failed trips")
                save_sample_df = pipeline.get_table(sample_table_name)
                save_sample_df.drop(trips_df.index, level="trip_id", inplace=True)
                pipeline.replace_table(sample_table_name, save_sample_df)
                del save_sample_df

    # if we estimated trip_destination, there should have been no failed trips
    # if we didn't, but it is enabled, it is probably a configuration error
    # if we just estimated trip_purpose, it isn't clear what they are trying to do , nor how to handle it
    assert not (
        estimation.manager.begin_estimation("trip_purpose")
        or estimation.manager.begin_estimation("trip_destination")
    )

    processed_trips = []
    save_samples = []
    i = 0
    TRIP_RESULT_COLUMNS = ["purpose", "destination", "origin", "failed"]
    while True:

        i += 1

        for c in TRIP_RESULT_COLUMNS:
            if c in trips_df:
                del trips_df[c]

        trips_df, save_sample_df = run_trip_purpose_and_destination(
            trips_df,
            tours_merged_df,
            chunk_size=chunk_size,
            trace_hh_id=trace_hh_id,
            trace_label=tracing.extend_trace_label(trace_label, "i%s" % i),
        )

        # # if testing, make sure at least one trip fails
        if (
            config.setting("testing_fail_trip_destination", False)
            and (i == 1)
            and not trips_df.failed.any()
        ):
            fail_o = trips_df[trips_df.trip_num < trips_df.trip_count].origin.max()
            trips_df.failed = (trips_df.origin == fail_o) & (
                trips_df.trip_num < trips_df.trip_count
            )

        num_failed_trips = trips_df.failed.sum()

        # if there were no failed trips, we are done
        if num_failed_trips == 0:
            processed_trips.append(trips_df[TRIP_RESULT_COLUMNS])
            if save_sample_df is not None:
                save_samples.append(save_sample_df)
            break

        logger.warning(
            "%s %s failed trips in iteration %s" % (trace_label, num_failed_trips, i)
        )
        file_name = "%s_i%s_failed_trips" % (trace_label, i)
        logger.info("writing failed trips to %s" % file_name)
        tracing.write_csv(
            trips_df[trips_df.failed], file_name=file_name, transpose=False
        )

        # if max iterations reached, add remaining trips to processed_trips and give up
        # note that we do this BEFORE failing leg_mates so resulting trip legs are complete
        if i >= MAX_ITERATIONS:
            logger.warning("%s too many iterations %s" % (trace_label, i))
            processed_trips.append(trips_df[TRIP_RESULT_COLUMNS])
            if save_sample_df is not None:
                save_sample_df.drop(
                    trips_df[trips_df.failed].index, level="trip_id", inplace=True
                )
                save_samples.append(save_sample_df)
            break

        # otherwise, if any trips failed, then their leg-mates trips must also fail
        flag_failed_trip_leg_mates(trips_df, "failed")

        # add the good trips to processed_trips
        processed_trips.append(trips_df[~trips_df.failed][TRIP_RESULT_COLUMNS])

        # and keep the failed ones to retry
        trips_df = trips_df[trips_df.failed]
        tours_merged_df = tours_merged_df[tours_merged_df.index.isin(trips_df.tour_id)]

        #  add trip samples of processed_trips to processed_samples
        if save_sample_df is not None:
            # drop failed trip samples
            save_sample_df.drop(trips_df.index, level="trip_id", inplace=True)
            save_samples.append(save_sample_df)

    # - assign result columns to trips
    processed_trips = pd.concat(processed_trips)

    if len(save_samples) > 0:
        save_sample_df = pd.concat(save_samples)
        logger.info(
            "adding %s samples to %s" % (len(save_sample_df), sample_table_name)
        )
        pipeline.extend_table(sample_table_name, save_sample_df)

    logger.info(
        "%s %s failed trips after %s iterations"
        % (trace_label, processed_trips.failed.sum(), i)
    )

    trips_df = trips.to_frame()
    assign_in_place(trips_df, processed_trips)

    trips_df = cleanup_failed_trips(trips_df)

    pipeline.replace_table("trips", trips_df)

    # check to make sure we wrote sample file if requestsd
    if want_sample_table and len(trips_df) > 0:
        assert pipeline.is_table(sample_table_name)
        # since we have saved samples for all successful trips
        # once we discard failed trips, we should samples for all trips
        save_sample_df = pipeline.get_table(sample_table_name)
        # expect samples only for intermediate trip destinatinos
        assert len(save_sample_df.index.get_level_values(0).unique()) == len(
            trips_df[trips_df.trip_num < trips_df.trip_count]
        )
        del save_sample_df

    if trace_hh_id:
        tracing.trace_df(
            trips_df,
            label=trace_label,
            slicer="trip_id",
            index_label="trip_id",
            warn_if_empty=True,
        )



import os
from pathlib import Path

import numpy as np
import pandas as pd
import yaml
from larch import DataFrames, Model
from larch.util import Dict

from .general import (
    apply_coefficients,
    construct_nesting_tree,
    dict_of_linear_utility_from_spec,
    remove_apostrophes,
)


def construct_availability(model, chooser_data, alt_codes_to_names):
    """
    Construct an availability dataframe based on -999 parameters.

    Parameters
    ----------
    model : larch.Model
    chooser_data : pandas.DataFrame
    alt_codes_to_names : Mapping[int,str]

    Returns
    -------
    pandas.DataFrame
    """
    avail = {}
    for acode, aname in alt_codes_to_names.items():
        unavail_cols = list(
            (
                chooser_data[i.data]
                if i.data in chooser_data
                else chooser_data.eval(i.data)
            )
            for i in model.utility_co[acode]
            if i.param == "-999"
        )
        if len(unavail_cols):
            avail[acode] = sum(unavail_cols) == 0
        else:
            avail[acode] = 1
    avail = pd.DataFrame(avail).astype(np.int8)
    avail.index = chooser_data.index
    return avail


def simple_simulate_data(
    name="tour_mode_choice",
    edb_directory="output/estimation_data_bundle/{name}/",
    coefficients_file="{name}_coefficients.csv",
    coefficients_template="{name}_coefficients_template.csv",
    spec_file="{name}_SPEC.csv",
    settings_file="{name}_model_settings.yaml",
    chooser_data_file="{name}_values_combined.csv",
    values_index_col="tour_id",
):
    edb_directory = edb_directory.format(name=name)

    def _read_csv(filename, **kwargs):
        filename = filename.format(name=name)
        return pd.read_csv(os.path.join(edb_directory, filename), **kwargs)

    settings_file = settings_file.format(name=name)
    with open(os.path.join(edb_directory, settings_file), "r") as yf:
        settings = yaml.load(
            yf,
            Loader=yaml.SafeLoader,
        )

    try:
        coefficients = _read_csv(
            coefficients_file,
            index_col="coefficient_name",
        )

        try:
            coef_template = _read_csv(
                coefficients_template,
                index_col="coefficient_name",
            )
        except FileNotFoundError:
            coef_template = None

        spec = _read_csv(spec_file, comment="#")
        spec = remove_apostrophes(spec, ["Label"])

        # remove temp rows from spec, ASim uses them to calculate the other values written
        # to the EDB, but they are not actually part of the utility function themselves.
        spec = spec.loc[~spec.Expression.isna()]
        spec = spec.loc[~spec.Expression.str.startswith("_")].copy()

        alt_names = list(spec.columns[3:])
        alt_codes = np.arange(1, len(alt_names) + 1)
        alt_names_to_codes = dict(zip(alt_names, alt_codes))
        alt_codes_to_names = dict(zip(alt_codes, alt_names))

        chooser_data = _read_csv(
            chooser_data_file,
            index_col=values_index_col,
        )

    except Exception:
        # when an error happens in reading anything other than settings, print settings
        from pprint import pprint

        pprint(settings)
        raise

    return Dict(
        edb_directory=Path(edb_directory),
        settings=settings,
        chooser_data=chooser_data,
        coefficients=coefficients,
        coef_template=coef_template,
        spec=spec,
        alt_names=alt_names,
        alt_codes=alt_codes,
        alt_names_to_codes=alt_names_to_codes,
        alt_codes_to_names=alt_codes_to_names,
    )


def simple_simulate_model(
    name,
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
    choices=None,
    construct_avail=False,
    values_index_col="household_id",
):
    data = simple_simulate_data(
        name=name,
        edb_directory=edb_directory,
        values_index_col=values_index_col,
    )
    coefficients = data.coefficients
    # coef_template = data.coef_template # not used
    spec = data.spec
    chooser_data = data.chooser_data
    settings = data.settings

    alt_names = data.alt_names
    alt_codes = data.alt_codes

    from .general import clean_values

    chooser_data = clean_values(
        chooser_data,
        alt_names_to_codes=choices or data.alt_names_to_codes,
        choice_code="override_choice_code",
    )

    if settings.get("LOGIT_TYPE") == "NL":
        tree = construct_nesting_tree(data.alt_names, settings["NESTS"])
        m = Model(graph=tree)
    else:
        m = Model(alts=data.alt_codes_to_names)

    m.utility_co = dict_of_linear_utility_from_spec(
        spec,
        "Label",
        dict(zip(alt_names, alt_codes)),
    )

    apply_coefficients(coefficients, m)

    if construct_avail:
        avail = construct_availability(m, chooser_data, data.alt_codes_to_names)
    else:
        avail = True

    d = DataFrames(
        co=chooser_data,
        av=avail,
        alt_codes=alt_codes,
        alt_names=alt_names,
    )

    m.dataservice = d
    m.choice_co_code = "override_choice_code"

    if return_data:
        return (
            m,
            Dict(
                edb_directory=data.edb_directory,
                chooser_data=chooser_data,
                coefficients=coefficients,
                spec=spec,
                alt_names=alt_names,
                alt_codes=alt_codes,
                settings=settings,
            ),
        )

    return m


def auto_ownership_model(
    name="auto_ownership",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return simple_simulate_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
        choices={i: i + 1 for i in range(5)},  # choices are coded in data as integers,
        # not 'cars0' etc as appears in the spec
    )


def free_parking_model(
    name="free_parking",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return simple_simulate_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
        choices={
            True: 1,
            False: 2,
        },  # True is free parking, False is paid parking, names match spec positions
    )


def mandatory_tour_frequency_model(
    name="mandatory_tour_frequency",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return simple_simulate_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
    )


def joint_tour_frequency_model(
    name="joint_tour_frequency",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return simple_simulate_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
    )


def atwork_subtour_frequency_model(
    name="atwork_subtour_frequency",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return simple_simulate_model(
        name=name,
        edb_directory=edb_directory,
        values_index_col="tour_id",
        return_data=return_data,
    )


def joint_tour_composition_model(
    name="joint_tour_composition",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return simple_simulate_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
    )


def joint_tour_participation_model(
    name="joint_tour_participation",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return simple_simulate_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
        values_index_col="participant_id",
        choices={
            0: 1,  # 0 means participate, alternative 1
            1: 2,  # 1 means not participate, alternative 2
        },
    )



# ActivitySim
# See full license in LICENSE.txt.

import argparse
import os
import sys

import extensions

from activitysim.cli.run import add_run_args, run

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    add_run_args(parser)
    args = parser.parse_args()

    os.environ["MKL_NUM_THREADS"] = "1"

    sys.exit(run(args))



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate
from activitysim.core import timetable as tt
from activitysim.core import tracing
from activitysim.core.util import assign_in_place

from .util import estimation
from .util.tour_scheduling import run_tour_scheduling
from .util.vectorize_tour_scheduling import vectorize_tour_scheduling

logger = logging.getLogger(__name__)
DUMP = False


@inject.step()
def non_mandatory_tour_scheduling(
    tours, persons_merged, tdd_alts, chunk_size, trace_hh_id
):
    """
    This model predicts the departure time and duration of each activity for non-mandatory tours
    """

    model_name = "non_mandatory_tour_scheduling"
    trace_label = model_name

    persons_merged = persons_merged.to_frame()

    tours = tours.to_frame()
    non_mandatory_tours = tours[tours.tour_category == "non_mandatory"]

    # - if no mandatory_tours
    if non_mandatory_tours.shape[0] == 0:
        tracing.no_results(model_name)
        return

    tour_segment_col = None

    choices = run_tour_scheduling(
        model_name,
        non_mandatory_tours,
        persons_merged,
        tdd_alts,
        tour_segment_col,
        chunk_size,
        trace_hh_id,
    )

    assign_in_place(tours, choices)
    pipeline.replace_table("tours", tours)

    # updated df for tracing
    non_mandatory_tours = tours[tours.tour_category == "non_mandatory"]

    tracing.dump_df(
        DUMP,
        tt.tour_map(persons_merged, non_mandatory_tours, tdd_alts),
        trace_label,
        "tour_map",
    )

    if trace_hh_id:
        tracing.trace_df(
            non_mandatory_tours,
            label=trace_label,
            slicer="person_id",
            index_label="tour_id",
            columns=None,
            warn_if_empty=True,
        )



# ActivitySim
# See full license in LICENSE.txt.
import logging
import os

import numpy as np
import pandas as pd

from activitysim.core import config, inject
from activitysim.core import timetable as tt

logger = logging.getLogger(__name__)


@inject.injectable(cache=True)
def tdd_alts():
    # right now this file just contains the start and end hour
    file_path = config.config_file_path("tour_departure_and_duration_alternatives.csv")
    df = pd.read_csv(file_path)

    df["duration"] = df.end - df.start

    # - NARROW
    df = df.astype(np.int8)

    return df


@inject.injectable(cache=True)
def tdd_alt_segments():

    # tour_purpose,time_period,start,end
    # work,EA,3,5
    # work,AM,6,8
    # ...
    # school,PM,15,17
    # school,EV,18,22

    file_path = config.config_file_path(
        "tour_departure_and_duration_segments.csv", mandatory=False
    )

    if file_path:

        df = pd.read_csv(file_path, comment="#")

        # - NARROW
        df["start"] = df["start"].astype(np.int8)
        df["end"] = df["end"].astype(np.int8)

    else:
        df = None

    return df


@inject.table()
def person_windows(persons, tdd_alts):

    df = tt.create_timetable_windows(persons, tdd_alts)

    inject.add_table("person_windows", df)

    return df


@inject.injectable()
def timetable(person_windows, tdd_alts):

    logging.debug("@inject timetable")
    return tt.TimeTable(person_windows.to_frame(), tdd_alts, person_windows.name)



# ActivitySim
# See full license in LICENSE.txt.

from builtins import range

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from .. import chunk
from .. import timetable as tt


@pytest.fixture
def persons():

    df = pd.DataFrame(index=list(range(6)))

    return df


@pytest.fixture
def tdd_alts():
    alts = pd.DataFrame(
        data=[
            [5, 5],
            [5, 6],
            [5, 7],
            [5, 8],
            [5, 9],
            [5, 10],
            [6, 6],
            [6, 7],
            [6, 8],
            [6, 9],
            [6, 10],
            [7, 7],
            [7, 8],
            [7, 9],
            [7, 10],
            [8, 8],
            [8, 9],
            [8, 10],
            [9, 9],
            [9, 10],
            [10, 10],
        ],
        columns=["start", "end"],
    )
    alts["duration"] = alts.end - alts.start

    return alts


def test_basic(persons, tdd_alts):

    with chunk.chunk_log("test_basic", base=True):

        person_windows = tt.create_timetable_windows(persons, tdd_alts)

        timetable = tt.TimeTable(person_windows, tdd_alts, "person_windows")

        # print "\ntdd_footprints_df\n", timetable.tdd_footprints_df
        #     0  1  2  3  4  5  6  7
        # 0   0  6  0  0  0  0  0  0
        # 1   0  2  4  0  0  0  0  0
        # 2   0  2  7  4  0  0  0  0
        # 3   0  2  7  7  4  0  0  0
        # 4   0  2  7  7  7  4  0  0
        # 5   0  2  7  7  7  7  4  0
        # 6   0  0  6  0  0  0  0  0
        # 7   0  0  2  4  0  0  0  0
        # 8   0  0  2  7  4  0  0  0
        # 9   0  0  2  7  7  4  0  0
        # 10  0  0  2  7  7  7  4  0
        # 11  0  0  0  6  0  0  0  0
        # 12  0  0  0  2  4  0  0  0
        # 13  0  0  0  2  7  4  0  0
        # 14  0  0  0  2  7  7  4  0
        # 15  0  0  0  0  6  0  0  0
        # 16  0  0  0  0  2  4  0  0
        # 17  0  0  0  0  2  7  4  0
        # 18  0  0  0  0  0  6  0  0
        # 19  0  0  0  0  0  2  4  0
        # 20  0  0  0  0  0  0  6  0

        num_alts = len(tdd_alts.index)
        num_persons = len(persons.index)

        person_ids = pd.Series(list(range(num_persons)) * num_alts)
        tdds = pd.Series(np.repeat(list(range(num_alts)), num_persons))

        assert timetable.tour_available(person_ids, tdds).all()

        person_ids = pd.Series([0, 1, 2, 3, 4, 5])
        tdds = pd.Series([0, 1, 2, 15, 16, 17])
        timetable.assign(person_ids, tdds)

        # print "\nupdated_person_windows\n", timetable.get_person_windows_df()
        #    4  5  6  7  8  9  10  11
        # 0  0  6  0  0  0  0   0   0
        # 1  0  2  4  0  0  0   0   0
        # 2  0  2  7  4  0  0   0   0
        # 3  0  0  0  0  6  0   0   0
        # 4  0  0  0  0  2  4   0   0
        # 5  0  0  0  0  2  7   4   0

        person_ids = pd.Series([0, 1, 1, 0, 1, 3, 4])
        tdds = pd.Series(
            [
                0,  # tdd START_END does not collide with START_END
                0,  # tdd START_END does not collide with START
                6,  # tdd START_END does not collide with END
                1,  # tdd START does not collide with START_END
                7,  # tdd START does not collide with END
                3,  # tdd END does not collide with START_END
                3,  # tdd END does not collide with START
            ]
        )
        assert timetable.tour_available(person_ids, tdds).all()

        # print "\nupdated_person_windows\n", timetable.get_person_windows_df()
        #    4  5  6  7  8  9  10  11
        # 0  0  6  0  0  0  0   0   0
        # 1  0  2  4  0  0  0   0   0
        # 2  0  2  7  4  0  0   0   0
        # 3  0  0  0  0  6  0   0   0
        # 4  0  0  0  0  2  4   0   0
        # 5  0  0  0  0  2  7   4   0

        person_ids = pd.Series([1, 5, 2, 2])
        tdds = pd.Series(
            [
                1,  # tdd START + END collides with START + END
                17,  # START + MIDDLE + END collides with same
                6,  # tdd START_END collides with MIDDLE
                1,  # tdd START + END collides with START + MIDDLE
            ]
        )
        assert not timetable.tour_available(person_ids, tdds).any()

        # ensure that tour_available handles heterogeneous results
        person_ids = pd.Series([0, 1, 1, 5])
        tdds = pd.Series(
            [
                0,  # tdd START_END does not collide with START_END
                0,  # tdd START_END does not collide with START
                1,  # tdd START + END collides with START + END
                17,  # START + MIDDLE + END collides with same
            ]
        )
        pdt.assert_series_equal(
            timetable.tour_available(person_ids, tdds),
            pd.Series([True, True, False, False], index=person_ids.index),
        )

        # assigning overlapping trip END,START should convert END to START_END
        person_ids = pd.Series([2])
        tdds = pd.Series([13])
        assert timetable.tour_available(person_ids, tdds).all()
        assert timetable.windows[2, 3] == tt.I_END
        timetable.assign(person_ids, tdds)
        assert timetable.windows[2, 3] == tt.I_START_END

        # print "\nupdated_person_windows\n", timetable.get_person_windows_df()
        #    4  5  6  7  8  9  10  11
        # 0  0  6  0  0  0  0   0   0
        # 1  0  2  4  0  0  0   0   0
        # 2  0  2  7  6  7  4   0   0
        # 3  0  0  0  0  6  0   0   0
        # 4  0  0  0  0  2  4   0   0
        # 5  0  0  0  0  2  7   4   0

        # - previous_tour_ends

        person_ids = pd.Series([0, 1, 2, 3, 4, 5, 2])
        periods = pd.Series([5, 6, 9, 8, 9, 10, 7])
        assert timetable.previous_tour_ends(person_ids, periods).all()

        person_ids = pd.Series([0, 1, 2])
        periods = pd.Series([9, 5, 8])
        assert not timetable.previous_tour_ends(person_ids, periods).any()

        # - previous_tour_begins

        person_ids = pd.Series([0, 1, 2, 3, 4, 5, 2])
        periods = pd.Series([5, 5, 5, 8, 8, 8, 7])
        assert timetable.previous_tour_begins(person_ids, periods).all()

        person_ids = pd.Series([0, 1, 2])
        periods = pd.Series([9, 6, 8])
        assert not timetable.previous_tour_begins(person_ids, periods).any()

        # - adjacent_window_after
        person_ids = pd.Series([0, 1, 2, 3, 4, 5])
        periods = pd.Series([5, 5, 5, 5, 5, 5])
        adjacent_run_length = timetable.adjacent_window_after(person_ids, periods)
        pdt.assert_series_equal(adjacent_run_length, pd.Series([5, 5, 0, 5, 5, 3]))

        # - adjacent_window_before
        person_ids = pd.Series([0, 1, 2, 3, 4, 5])
        periods = pd.Series([10, 10, 10, 10, 10, 10])
        adjacent_run_length = timetable.adjacent_window_before(person_ids, periods)
        pdt.assert_series_equal(adjacent_run_length, pd.Series([5, 5, 1, 5, 5, 0]))

        # - remaining_periods_available
        person_ids = pd.Series([0, 1, 2, 3])
        starts = pd.Series([9, 6, 9, 5])
        ends = pd.Series([10, 10, 10, 9])
        periods_available = timetable.remaining_periods_available(
            person_ids, starts, ends
        )
        pdt.assert_series_equal(
            periods_available, pd.Series([6, 3, 4, 3]), check_dtype=False
        )



# ActivitySim
# See full license in LICENSE.txt.
import argparse
import logging
import os
import sys
import warnings

import numpy as np

from activitysim.core import chunk, config, inject, mem, pipeline, tracing

logger = logging.getLogger(__name__)


INJECTABLES = ["data_dir", "configs_dir", "output_dir", "settings_file_name"]


def add_run_args(parser, multiprocess=True):
    """Run command args"""
    parser.add_argument(
        "-w",
        "--working_dir",
        type=str,
        metavar="PATH",
        help="path to example/project directory (default: %s)" % os.getcwd(),
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        action="append",
        metavar="PATH",
        help="path to config dir",
    )
    parser.add_argument(
        "-o", "--output", type=str, metavar="PATH", help="path to output dir"
    )
    parser.add_argument(
        "-d",
        "--data",
        type=str,
        action="append",
        metavar="PATH",
        help="path to data dir",
    )
    parser.add_argument(
        "-r", "--resume", type=str, metavar="STEPNAME", help="resume after step"
    )
    parser.add_argument(
        "-p", "--pipeline", type=str, metavar="FILE", help="pipeline file name"
    )
    parser.add_argument(
        "-s", "--settings_file", type=str, metavar="FILE", help="settings file name"
    )
    parser.add_argument(
        "-g", "--chunk_size", type=int, metavar="BYTES", help="chunk size"
    )
    parser.add_argument(
        "--chunk_training_mode",
        type=str,
        help="chunk training mode, one of [training, adaptive, production, disabled]",
    )
    parser.add_argument(
        "--households_sample_size", type=int, metavar="N", help="households sample size"
    )

    if multiprocess:
        parser.add_argument(
            "-m",
            "--multiprocess",
            default=False,
            const=-1,
            metavar="(N)",
            nargs="?",
            type=int,
            help="run multiprocess. Adds configs_mp settings"
            " by default. Optionally give a number of processes,"
            " which will override the settings file.",
        )


def validate_injectable(name):
    try:
        dir_paths = inject.get_injectable(name)
    except RuntimeError:
        # injectable is missing, meaning is hasn't been explicitly set
        # and defaults cannot be found.
        sys.exit(
            "Error: please specify either a --working_dir "
            "containing 'configs', 'data', and 'output' folders "
            "or all three of --config, --data, and --output"
        )

    dir_paths = [dir_paths] if isinstance(dir_paths, str) else dir_paths

    for dir_path in dir_paths:
        if not os.path.exists(dir_path):
            sys.exit("Could not find %s '%s'" % (name, os.path.abspath(dir_path)))

    return dir_paths


def handle_standard_args(args, multiprocess=True):
    def inject_arg(name, value, cache=False):
        assert name in INJECTABLES
        inject.add_injectable(name, value, cache=cache)

    if args.working_dir:
        # activitysim will look in the current working directory for
        # 'configs', 'data', and 'output' folders by default
        os.chdir(args.working_dir)

    # settings_file_name should be cached or else it gets squashed by config.py
    if args.settings_file:
        inject_arg("settings_file_name", args.settings_file, cache=True)

    if args.config:
        inject_arg("configs_dir", args.config)

    if args.data:
        inject_arg("data_dir", args.data)

    if args.output:
        inject_arg("output_dir", args.output)

    if multiprocess and args.multiprocess:
        config_paths = validate_injectable("configs_dir")

        if not os.path.exists("configs_mp"):
            logger.warning("could not find 'configs_mp'. skipping...")
        else:
            logger.info("adding 'configs_mp' to config_dir list...")
            config_paths.insert(0, "configs_mp")
            inject_arg("configs_dir", config_paths)

        config.override_setting("multiprocess", True)
        if args.multiprocess > 0:
            config.override_setting("num_processes", args.multiprocess)

    if args.chunk_size:
        config.override_setting("chunk_size", int(args.chunk_size))
    if args.chunk_training_mode is not None:
        config.override_setting("chunk_training_mode", args.chunk_training_mode)
    if args.households_sample_size is not None:
        config.override_setting("households_sample_size", args.households_sample_size)

    for injectable in ["configs_dir", "data_dir", "output_dir"]:
        validate_injectable(injectable)

    if args.pipeline:
        inject.add_injectable("pipeline_file_name", args.pipeline)

    if args.resume:
        config.override_setting("resume_after", args.resume)


def cleanup_output_files():

    tracing.delete_trace_files()

    tracing.delete_output_files("h5")
    tracing.delete_output_files("csv")
    tracing.delete_output_files("txt")
    tracing.delete_output_files("yaml")
    tracing.delete_output_files("prof")
    tracing.delete_output_files("omx")


def run(args):
    """
    Run the models. Specify a project folder using the '--working_dir' option,
    or point to the config, data, and output folders directly with
    '--config', '--data', and '--output'. Both '--config' and '--data' can be
    specified multiple times. Directories listed first take precedence.

    returns:
        int: sys.exit exit code
    """

    # register abm steps and other abm-specific injectables
    # by default, assume we are running activitysim.abm
    # other callers (e.g. populationsim) will have to arrange to register their own steps and injectables
    # (presumably) in a custom run_simulation.py instead of using the 'activitysim run' command
    if not inject.is_injectable("preload_injectables"):
        from activitysim import (  # register abm steps and other abm-specific injectables
            abm,
        )

    tracing.config_logger(basic=True)
    handle_standard_args(args)  # possibly update injectables

    # legacy support for run_list setting nested 'models' and 'resume_after' settings
    if config.setting("run_list"):
        warnings.warn(
            "Support for 'run_list' settings group will be removed.\n"
            "The run_list.steps setting is renamed 'models'.\n"
            "The run_list.resume_after setting is renamed 'resume_after'.\n"
            "Specify both 'models' and 'resume_after' directly in settings config file.",
            FutureWarning,
        )
        run_list = config.setting("run_list")
        if "steps" in run_list:
            assert not config.setting(
                "models"
            ), f"Don't expect 'steps' in run_list and 'models' as stand-alone setting!"
            config.override_setting("models", run_list["steps"])

        if "resume_after" in run_list:
            assert not config.setting(
                "resume_after"
            ), f"Don't expect 'resume_after' both in run_list and as stand-alone setting!"
            config.override_setting("resume_after", run_list["resume_after"])

    # If you provide a resume_after argument to pipeline.run
    # the pipeline manager will attempt to load checkpointed tables from the checkpoint store
    # and resume pipeline processing on the next submodel step after the specified checkpoint
    resume_after = config.setting("resume_after", None)

    # cleanup if not resuming
    if not resume_after:
        cleanup_output_files()
    elif config.setting("cleanup_trace_files_on_resume", False):
        tracing.delete_trace_files()

    tracing.config_logger(basic=False)  # update using possibly new logging configs
    config.filter_warnings()
    logging.captureWarnings(capture=True)

    # directories
    for k in ["configs_dir", "settings_file_name", "data_dir", "output_dir"]:
        logger.info("SETTING %s: %s" % (k, inject.get_injectable(k, None)))

    log_settings = inject.get_injectable("log_settings", {})
    for k in log_settings:
        logger.info("SETTING %s: %s" % (k, config.setting(k)))

    # OMP_NUM_THREADS: openmp
    # OPENBLAS_NUM_THREADS: openblas
    # MKL_NUM_THREADS: mkl
    for env in ["MKL_NUM_THREADS", "OMP_NUM_THREADS", "OPENBLAS_NUM_THREADS"]:
        logger.info(f"ENV {env}: {os.getenv(env)}")

    np_info_keys = [
        "atlas_blas_info",
        "atlas_blas_threads_info",
        "atlas_info",
        "atlas_threads_info",
        "blas_info",
        "blas_mkl_info",
        "blas_opt_info",
        "lapack_info",
        "lapack_mkl_info",
        "lapack_opt_info",
        "mkl_info",
    ]

    for cfg_key in np_info_keys:
        info = np.__config__.get_info(cfg_key)
        if info:
            for info_key in ["libraries"]:
                if info_key in info:
                    logger.info(f"NUMPY {cfg_key} {info_key}: {info[info_key]}")

    t0 = tracing.print_elapsed_time()

    try:
        if config.setting("multiprocess", False):
            logger.info("run multiprocess simulation")

            from activitysim.core import mp_tasks

            injectables = {k: inject.get_injectable(k) for k in INJECTABLES}
            mp_tasks.run_multiprocess(injectables)

            assert not pipeline.is_open()

            if config.setting("cleanup_pipeline_after_run", False):
                pipeline.cleanup_pipeline()

        else:
            logger.info("run single process simulation")

            pipeline.run(models=config.setting("models"), resume_after=resume_after)

            if config.setting("cleanup_pipeline_after_run", False):
                pipeline.cleanup_pipeline()  # has side effect of closing open pipeline
            else:
                pipeline.close_pipeline()

            mem.log_global_hwm()  # main process
    except Exception:
        # log time until error and the error traceback
        tracing.print_elapsed_time("all models until this error", t0)
        logger.exception("activitysim run encountered an unrecoverable error")
        raise

    chunk.consolidate_logs()
    mem.consolidate_logs()

    tracing.print_elapsed_time("all models", t0)

    return 0


if __name__ == "__main__":

    from activitysim import abm  # register injectables

    parser = argparse.ArgumentParser()
    add_run_args(parser)
    args = parser.parse_args()

    parser.parse_args(["--sum", "7", "-1", "42"])
    sys.exit(run(args))



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate, tracing

from .util import estimation
from .util.tour_frequency import process_atwork_subtours

logger = logging.getLogger(__name__)


def add_null_results(trace_label, tours):
    logger.info("Skipping %s: add_null_results", trace_label)
    tours["atwork_subtour_frequency"] = np.nan
    pipeline.replace_table("tours", tours)


@inject.step()
def atwork_subtour_frequency(tours, persons_merged, chunk_size, trace_hh_id):
    """
    This model predicts the frequency of making at-work subtour tours
    (alternatives for this model come from a separate csv file which is
    configured by the user).
    """

    trace_label = "atwork_subtour_frequency"
    model_settings_file_name = "atwork_subtour_frequency.yaml"

    tours = tours.to_frame()
    work_tours = tours[tours.tour_type == "work"]

    # - if no work_tours
    if len(work_tours) == 0:
        add_null_results(trace_label, tours)
        return

    model_settings = config.read_model_settings(model_settings_file_name)
    estimator = estimation.manager.begin_estimation("atwork_subtour_frequency")

    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    alternatives = simulate.read_model_alts(
        "atwork_subtour_frequency_alternatives.csv", set_index="alt"
    )

    # merge persons into work_tours
    persons_merged = persons_merged.to_frame()
    work_tours = pd.merge(
        work_tours, persons_merged, left_on="person_id", right_index=True
    )

    logger.info("Running atwork_subtour_frequency with %d work tours", len(work_tours))

    nest_spec = config.get_logit_model_settings(model_settings)
    constants = config.get_model_constants(model_settings)

    # - preprocessor
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:

        expressions.assign_columns(
            df=work_tours, model_settings=preprocessor_settings, trace_label=trace_label
        )

    if estimator:
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_choosers(work_tours)

    choices = simulate.simple_simulate(
        choosers=work_tours,
        spec=model_spec,
        nest_spec=nest_spec,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="atwork_subtour_frequency",
        estimator=estimator,
    )

    # convert indexes to alternative names
    choices = pd.Series(model_spec.columns[choices.values], index=choices.index)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(
            choices, "tours", "atwork_subtour_frequency"
        )
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    # add atwork_subtour_frequency column to tours
    # reindex since we are working with a subset of tours
    tours["atwork_subtour_frequency"] = choices.reindex(tours.index)
    pipeline.replace_table("tours", tours)

    # - create atwork_subtours based on atwork_subtour_frequency choice names
    work_tours = tours[tours.tour_type == "work"]
    assert not work_tours.atwork_subtour_frequency.isnull().any()

    subtours = process_atwork_subtours(work_tours, alternatives)

    tours = pipeline.extend_table("tours", subtours)

    tracing.register_traceable_table("tours", subtours)
    pipeline.get_rn_generator().add_channel("tours", subtours)

    tracing.print_summary(
        "atwork_subtour_frequency", tours.atwork_subtour_frequency, value_counts=True
    )

    if trace_hh_id:
        tracing.trace_df(tours, label="atwork_subtour_frequency.tours")



# crop marin tvpb example data processing to one county
# Ben Stabler, ben.stabler@rsginc.com, 09/17/20

import argparse
import os

import numpy as np
import openmatrix as omx
import pandas as pd

MAZ_OFFSET = 100000

segments = {
    "test": {"DistName": ["Downtown SF"]},
    "marin_sf": {"CountyName": ["Marin", "San Francisco"]},
    "full": {},
}

parser = argparse.ArgumentParser(description="crop Marin raw_data")
parser.add_argument(
    "segment_name",
    metavar="segment_name",
    type=str,
    nargs=1,
    help=f"geography segmentation (e.g. full)",
)

parser.add_argument(
    "-c",
    "--check_geography",
    default=False,
    action="store_true",
    help="check consistency of MAZ, TAZ, TAP zone_ids and foreign keys & write orphan_households file",
)

args = parser.parse_args()


segment_name = args.segment_name[0]
check_geography = args.check_geography

assert segment_name in segments.keys(), f"Unknown seg: {segment_name}"

input_dir = "./data_raw"
output_dir = f"./data_{segment_name}"


print(f"segment_name {segment_name}")

print(f"input_dir {input_dir}")
print(f"output_dir {output_dir}")

print(f"check_geography {check_geography}")

if not os.path.isdir(output_dir):
    print(f"creating output directory {output_dir}")
    os.mkdir(output_dir)


def input_path(file_name):
    return os.path.join(input_dir, file_name)


def output_path(file_name):
    return os.path.join(output_dir, file_name)


def patch_maz(df, maz_offset):
    for c in df.columns:
        if c in ["MAZ", "OMAZ", "DMAZ", "mgra", "orig_mgra", "dest_mgra"]:
            df[c] += maz_offset
    return df


def read_csv(file_name):
    df = pd.read_csv(input_path(file_name))
    if MAZ_OFFSET:
        df = patch_maz(df, MAZ_OFFSET)
    print(f"read {file_name} {df.shape}")
    return df


def to_csv(df, file_name):
    df.to_csv(output_path(file_name), index=False)
    print(f"write {file_name} {df.shape}")


# non-standard input file names
LAND_USE = "maz_data_asim.csv"
HOUSEHOLDS = "households_asim.csv"
PERSONS = "persons_asim.csv"
MAZ_TAZ = "maz_taz.csv"
TAP_MAZ = "tap_data.csv"
ACCESSIBILITY = "access.csv"
WORK_TOURS = "work_tours.csv"

if check_geography:

    # ######## check for orphan_households not in any maz in land_use
    land_use = read_csv(LAND_USE)
    land_use = land_use[["MAZ", "TAZ"]]
    land_use = land_use.sort_values(["TAZ", "MAZ"])

    households = read_csv(HOUSEHOLDS)
    orphan_households = households[~households.MAZ.isin(land_use.MAZ)]
    print(f"{len(orphan_households)} orphan_households")

    # write orphan_households to INPUT directory (since it doesn't belong in output)
    file_name = "orphan_households.csv"
    print(f"writing {file_name} {orphan_households.shape} to {input_path(file_name)}")
    orphan_households.to_csv(input_path(file_name), index=False)

    # ######## check that land_use and maz and taz tables have same MAZs and TAZs

    # could just build maz and taz files, but want to make sure PSRC data is right

    land_use = read_csv(LAND_USE)
    land_use = land_use.sort_values("MAZ")
    maz = read_csv(MAZ_TAZ).sort_values("MAZ")

    # ### FATAL ###
    if not land_use.MAZ.isin(maz.MAZ).all():
        print(
            f"land_use.MAZ not in maz.MAZ\n{land_use.MAZ[~land_use.MAZ.isin(maz.MAZ)]}"
        )
        raise RuntimeError(f"land_use.MAZ not in maz.MAZ")

    if not maz.MAZ.isin(land_use.MAZ).all():
        print(f"maz.MAZ not in land_use.MAZ\n{maz.MAZ[~maz.MAZ.isin(land_use.MAZ)]}")

    # ### FATAL ###
    if not land_use.TAZ.isin(maz.TAZ).all():
        print(
            f"land_use.TAZ not in maz.TAZ\n{land_use.TAZ[~land_use.TAZ.isin(maz.TAZ)]}"
        )
        raise RuntimeError(f"land_use.TAZ not in maz.TAZ")

    if not maz.TAZ.isin(land_use.TAZ).all():
        print(f"maz.TAZ not in land_use.TAZ\n{maz.TAZ[~maz.TAZ.isin(land_use.TAZ)]}")


# land_use

land_use = read_csv(LAND_USE)

slicer = segments[segment_name]
for slice_col, slice_values in slicer.items():
    print(f"slice {slice_col}: {slice_values}")

    land_use = land_use[land_use[slice_col].isin(slice_values)]

print(f"land_use shape after slicing {land_use.shape}")
to_csv(land_use, "land_use.csv")

# maz_taz, tazs, taps

maz_taz = land_use[["MAZ", "TAZ"]]
to_csv(maz_taz, "maz_taz.csv")

tazs = land_use["TAZ"].unique()
tazs.sort()

taps = read_csv(TAP_MAZ)
taps = taps[["TAP", "TAZ"]].sort_values(by="TAP")
taps = taps[taps["TAZ"].isin(tazs)]
to_csv(taps, "tap.csv")

# maz to tap walk, bike

maz_tap_walk = read_csv("maz_tap_walk.csv")
maz_maz_walk = read_csv("maz_maz_walk.csv")
maz_maz_bike = read_csv("maz_maz_bike.csv")

maz_tap_walk = maz_tap_walk[
    maz_tap_walk["MAZ"].isin(land_use["MAZ"]) & maz_tap_walk["TAP"].isin(taps["TAP"])
]
maz_maz_walk = maz_maz_walk[
    maz_maz_walk["OMAZ"].isin(land_use["MAZ"])
    & maz_maz_walk["DMAZ"].isin(land_use["MAZ"])
]
maz_maz_bike = maz_maz_bike[
    maz_maz_bike["OMAZ"].isin(land_use["MAZ"])
    & maz_maz_bike["DMAZ"].isin(land_use["MAZ"])
]

to_csv(maz_tap_walk, "maz_tap_walk.csv")
to_csv(maz_maz_walk, "maz_maz_walk.csv")
to_csv(maz_maz_bike, "maz_maz_bike.csv")

tap_lines = read_csv("tap_lines.csv")
tap_lines = tap_lines[tap_lines["TAP"].isin(taps["TAP"])]
to_csv(tap_lines, "tap_lines.csv")

# taz to tap drive data

taz_tap_drive = read_csv("maz_taz_tap_drive.csv")
taz_tap_drive = taz_tap_drive[
    taz_tap_drive["MAZ"].isin(land_use["MAZ"]) & taz_tap_drive["TAP"].isin(taps["TAP"])
]
to_csv(taz_tap_drive, "maz_taz_tap_drive.csv")


# accessibility data

access = read_csv(ACCESSIBILITY)
access = access[access["mgra"].isin(land_use["MAZ"])]
to_csv(access, "accessibility.csv")


# households

households = read_csv(HOUSEHOLDS)
households = households[households["MAZ"].isin(land_use["MAZ"])]
to_csv(households, "households.csv")

# persons

persons = read_csv(PERSONS)
persons = persons[persons["HHID"].isin(households["HHID"])]
to_csv(persons, "persons.csv")

# tours file

work_tours = read_csv(WORK_TOURS)
work_tours = work_tours[work_tours["hh_id"].isin(households["HHID"])]
work_tours = work_tours[
    work_tours["orig_mgra"].isin(land_use["MAZ"])
    & work_tours["dest_mgra"].isin(land_use["MAZ"])
]
to_csv(work_tours, "work_tours.csv")

# skims

taz_indexes = (tazs - 1).tolist()  # offset_map
tap_indexes = (taps["TAP"] - 1).tolist()  # offset_map
time_periods = ["AM", "EA", "EV", "MD", "PM"]
skim_data_type = np.float32

# taz skims with skim_data_type np.float32 are under 2GB - otherwise we would need to further segment them

for tp in time_periods:
    in_file_name = f"HWYSKM{tp}_taz_rename.omx"
    taz_file_in = omx.open_file(input_path(in_file_name))
    out_file_name = f"highway_skims_{tp}.omx"
    taz_file_out = omx.open_file(output_path(out_file_name), "w")
    taz_file_out.create_mapping("ZONE", tazs.tolist())
    for mat_name in taz_file_in.list_matrices():
        # make sure we have a vanilla numpy array, not a CArray
        m = np.asanyarray(taz_file_in[mat_name]).astype(skim_data_type)
        m = m[taz_indexes, :][:, taz_indexes]
        taz_file_out[mat_name] = m
        print(f"taz {mat_name} {m.shape}")
    taz_file_in.close()
    taz_file_out.close()

for skim_set in ["SET1", "SET2", "SET3"]:
    out_file_name = f"transit_skims_{skim_set}.omx"
    tap_file_out = omx.open_file(output_path(out_file_name), "w")
    tap_file_out.create_mapping("TAP", taps["TAP"].tolist())
    for tp in time_periods:
        in_file_name = f"transit_skims_{tp}_{skim_set}_rename.omx"
        tap_file_in = omx.open_file(input_path(in_file_name))
        for mat_name in tap_file_in.list_matrices():
            # make sure we have a vanilla numpy array, not a CArray
            m = np.asanyarray(tap_file_in[mat_name]).astype(skim_data_type)
            m = m[tap_indexes, :][:, tap_indexes]
            tap_file_out[mat_name] = m
            print(f"tap {skim_set} {mat_name} {m.shape}")
        tap_file_in.close()
    tap_file_out.close()



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np

from activitysim.abm.models.util import estimation
from activitysim.core import config, expressions, inject, pipeline, simulate, tracing

logger = logging.getLogger("activitysim")


@inject.step()
def work_from_home(persons_merged, persons, chunk_size, trace_hh_id):
    """
    This model predicts whether a person (worker) works from home. The output
    from this model is TRUE (if works from home) or FALSE (works away from home).
    """

    trace_label = "work_from_home"
    model_settings_file_name = "work_from_home.yaml"

    choosers = persons_merged.to_frame()
    model_settings = config.read_model_settings(model_settings_file_name)
    chooser_filter_column_name = model_settings.get("CHOOSER_FILTER_COLUMN_NAME")
    choosers = choosers[choosers[chooser_filter_column_name]]
    logger.info("Running %s with %d persons", trace_label, len(choosers))

    estimator = estimation.manager.begin_estimation("work_from_home")

    constants = config.get_model_constants(model_settings)
    work_from_home_alt = model_settings["WORK_FROM_HOME_ALT"]

    # - preprocessor
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:

        locals_d = {}
        if constants is not None:
            locals_d.update(constants)

        expressions.assign_columns(
            df=choosers,
            model_settings=preprocessor_settings,
            locals_dict=locals_d,
            trace_label=trace_label,
        )

    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)

    nest_spec = config.get_logit_model_settings(model_settings)

    if estimator:
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_spec(model_settings)
        estimator.write_coefficients(coefficients_df)
        estimator.write_choosers(choosers)

    # - iterative single process what-if adjustment if specified
    iterations = model_settings.get("WORK_FROM_HOME_ITERATIONS", 1)
    iterations_chooser_filter = model_settings.get(
        "WORK_FROM_HOME_CHOOSER_FILTER", None
    )
    iterations_coefficient_constant = model_settings.get(
        "WORK_FROM_HOME_COEFFICIENT_CONSTANT", None
    )
    iterations_target_percent = model_settings.get(
        "WORK_FROM_HOME_TARGET_PERCENT", None
    )
    iterations_target_percent_tolerance = model_settings.get(
        "WORK_FROM_HOME_TARGET_PERCENT_TOLERANCE", None
    )

    for iteration in range(iterations):

        logger.info(
            "Running %s with %d persons iteration %d",
            trace_label,
            len(choosers),
            iteration,
        )

        # re-read spec to reset substitution
        model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
        model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

        choices = simulate.simple_simulate(
            choosers=choosers,
            spec=model_spec,
            nest_spec=nest_spec,
            locals_d=constants,
            chunk_size=chunk_size,
            trace_label=trace_label,
            trace_choice_name="work_from_home",
            estimator=estimator,
        )

        if iterations_target_percent is not None:
            choices_for_filter = choices[choosers[iterations_chooser_filter]]

            current_percent = (choices_for_filter == work_from_home_alt).sum() / len(
                choices_for_filter
            )
            logger.info(
                "Running %s iteration %i choosers %i current percent %f target percent %f",
                trace_label,
                iteration,
                len(choices_for_filter),
                current_percent,
                iterations_target_percent,
            )

            if current_percent <= (
                iterations_target_percent + iterations_target_percent_tolerance
            ) and current_percent >= (
                iterations_target_percent - iterations_target_percent_tolerance
            ):
                logger.info(
                    "Running %s iteration %i converged with coefficient %f",
                    trace_label,
                    iteration,
                    coefficients_df.value[iterations_coefficient_constant],
                )
                break

            else:
                new_value = (
                    np.log(
                        iterations_target_percent / np.maximum(current_percent, 0.0001)
                    )
                    + coefficients_df.value[iterations_coefficient_constant]
                )
                coefficients_df.value[iterations_coefficient_constant] = new_value
                logger.info(
                    "Running %s iteration %i new coefficient for next iteration %f",
                    trace_label,
                    iteration,
                    new_value,
                )
                iteration = iteration + 1

    choices = choices == work_from_home_alt

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(choices, "persons", "work_from_home")
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    persons = persons.to_frame()
    persons["work_from_home"] = choices.reindex(persons.index).fillna(0).astype(bool)
    persons["is_out_of_home_worker"] = (
        persons[chooser_filter_column_name] & ~persons["work_from_home"]
    )

    pipeline.replace_table("persons", persons)

    tracing.print_summary("work_from_home", persons.work_from_home, value_counts=True)

    if trace_hh_id:
        tracing.trace_df(persons, label=trace_label, warn_if_empty=True)



# ActivitySim
# See full license in LICENSE.txt.
import logging
import os

import numpy as np
import pandas as pd

from activitysim.abm.models.trip_matrices import annotate_trips
from activitysim.core import config, expressions, inject, pipeline

logger = logging.getLogger(__name__)


def wrap_skims(
    network_los: pipeline.Pipeline, trips_merged: pd.DataFrame
) -> dict[str, object]:
    """
    Retrieve skim wrappers for merged trips.

    For each record in `trips_merged`, retrieve skim wrappers for appropriate time of day.

    Returns dictionary of skims wrappers that are available for use in expressions defined
    in `summarize_preprocessor.csv`
    """
    skim_dict = network_los.get_default_skim_dict()

    trips_merged["start_tour_period"] = network_los.skim_time_period_label(
        trips_merged["start"]
    )
    trips_merged["end_tour_period"] = network_los.skim_time_period_label(
        trips_merged["end"]
    )
    trips_merged["trip_period"] = network_los.skim_time_period_label(
        trips_merged["depart"]
    )

    tour_odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key="origin_tour",
        dest_key="destination_tour",
        dim3_key="start_tour_period",
    )
    tour_dot_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key="destination_tour", dest_key="origin_tour", dim3_key="end_tour_period"
    )
    trip_odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key="origin_trip", dest_key="destination_trip", dim3_key="trip_period"
    )

    tour_od_skim_stack_wrapper = skim_dict.wrap("origin_tour", "destination_tour")
    trip_od_skim_stack_wrapper = skim_dict.wrap("origin_trip", "destination_trip")

    return {
        "tour_odt_skims": tour_odt_skim_stack_wrapper,
        "tour_dot_skims": tour_dot_skim_stack_wrapper,
        "trip_odt_skims": trip_odt_skim_stack_wrapper,
        "tour_od_skims": tour_od_skim_stack_wrapper,
        "trip_od_skims": trip_od_skim_stack_wrapper,
    }


DEFAULT_BIN_LABEL_FORMAT = "{left:,.2f} - {right:,.2f}"


def construct_bin_labels(bins: pd.Series, label_format: str) -> pd.Series:
    """
    Construct bin label strings based on intervals (pd.Interval) in `bins`

    `label_format` is an F-string format that can reference the following variables:
     - 'left': Bin minimum
     - 'right': Min maximum
     - 'mid': Bin center
     - 'rank': Bin rank (lowest to highest)

     For example: '{left:,.2f} - {right:,.2f}' might yield '0.00 - 1.00'
    """
    left = bins.apply(lambda x: x.left)
    mid = bins.apply(lambda x: x.mid)
    right = bins.apply(lambda x: x.right)
    # Get integer ranks of bins (e.g., 1st, 2nd ... nth quantile)
    rank = mid.map(
        {
            x: sorted(mid.unique().tolist()).index(x) + 1 if pd.notnull(x) else np.nan
            for x in mid.unique()
        },
        na_action="ignore",
    )

    def construct_label(label_format, bounds_dict):
        bounds_dict = {
            x: bound for x, bound in bounds_dict.items() if x in label_format
        }
        return label_format.format(**bounds_dict)

    labels = pd.Series(
        [
            construct_label(
                label_format, {"left": lt, "mid": md, "right": rt, "rank": rk}
            )
            for lt, md, rt, rk in zip(left, mid, right, rank)
        ],
        index=bins.index,
    )
    # Convert to numeric if possible
    labels = pd.to_numeric(labels, errors="ignore")
    return labels


def quantiles(
    data: pd.Series, bins: pd.Series, label_format: str = DEFAULT_BIN_LABEL_FORMAT
) -> pd.Series:
    """
    Construct quantiles from a Series given a number of bins.

    For example: set bins = 5 to construct quintiles.

    data: Input Series
    bins: Number of bins
    label_format: F-string format for bin labels
        Bins are labeled with 'min - max' ranges by default.

    Returns a Series indexed by labels
    """
    vals = data.sort_values()
    # qcut a ranking instead of raw values to deal with high frequencies of the same value
    # (e.g., many 0 values) that may span multiple bins
    ranks = vals.rank(method="first")
    bins = pd.qcut(ranks, bins, duplicates="drop")
    bins = construct_bin_labels(bins, label_format)
    return bins


def spaced_intervals(
    data: pd.Series,
    lower_bound: float,
    interval: float,
    label_format: str = DEFAULT_BIN_LABEL_FORMAT,
) -> pd.Series:
    """
    Construct evenly-spaced intervals from a Series given a starting value and bin size.

    data: Input Series
    lower_bound: Minimum value of lowest bin
    interval: Bin spacing above the `lower_bound`
    label_format: F-string format for bin labels
        Bins are labeled with 'min - max' ranges by default.

    Returns a Series indexed by labels
    """
    if lower_bound == "min":
        lower_bound = data.min()
    breaks = np.arange(lower_bound, data.max() + interval, interval)
    bins = pd.cut(data, breaks, include_lowest=True)
    bins = construct_bin_labels(bins, label_format)
    return bins


def equal_intervals(
    data: pd.Series, bins: int, label_format: str = DEFAULT_BIN_LABEL_FORMAT
) -> pd.Series:
    """
    Construct equally-spaced intervals across the entire range of a Series.

    data: Input Series
    bins: Number of bins
    label_format: F-string format for bin labels
        Bins are labeled with 'min - max' ranges by default.

    Returns a Series indexed by labels
    """
    bins = pd.cut(data, bins, include_lowest=True)
    bins = construct_bin_labels(bins, label_format)
    return bins


def manual_breaks(
    data: pd.Series,
    bin_breaks: list,
    labels: list = None,
    label_format: str = DEFAULT_BIN_LABEL_FORMAT,
) -> pd.Series:
    """
    Classify numeric data in a Pandas Series into manually-defined bins.

    data: Input Series
    bin_breaks: Break points between bins
    labels: Manually-defined labels for each bin (`len(labels)` == `len(bin_breaks) + 1`)
    label_format: F-string format for bin labels if not defined by `labels`
        Bins are labeled with 'min - max' ranges by default.

    Returns a Series indexed by labels
    """
    if isinstance(labels, list):
        return pd.cut(data, bin_breaks, labels=labels, include_lowest=True)
    else:
        bins = pd.cut(data, bin_breaks, include_lowest=True)
        bins = construct_bin_labels(bins, label_format)
        return bins


@inject.step()
def summarize(
    network_los: pipeline.Pipeline,
    persons: pd.DataFrame,
    persons_merged: pd.DataFrame,
    households: pd.DataFrame,
    households_merged: pd.DataFrame,
    trips: pd.DataFrame,
    tours: pd.DataFrame,
    tours_merged: pd.DataFrame,
    land_use: pd.DataFrame,
):
    """
    A standard model that uses expression files to summarize pipeline tables for vizualization.

    Summaries are configured in `summarize.yaml`, including specification of the
    expression file (`summarize.csv` by default).

    Columns in pipeline tables can also be sliced and aggregated prior to summarization.
    This preprocessing is configured in `summarize.yaml`.

    Outputs a seperate csv summary file for each expression;
    outputs starting with '_' are saved as temporary local variables.
    """
    trace_label = "summarize"
    model_settings_file_name = "summarize.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    output_location = (
        model_settings["OUTPUT"] if "OUTPUT" in model_settings else "summaries"
    )
    os.makedirs(config.output_file_path(output_location), exist_ok=True)

    spec = pd.read_csv(
        config.config_file_path(model_settings["SPECIFICATION"]), comment="#"
    )

    # Load dataframes from pipeline
    persons = persons.to_frame()
    persons_merged = persons_merged.to_frame()
    households = households.to_frame()
    households_merged = households_merged.to_frame()
    trips = trips.to_frame()
    tours = tours_merged.to_frame()
    tours_merged = tours_merged.to_frame()
    land_use = land_use.to_frame()

    # - trips_merged - merge trips and tours_merged
    trips_merged = pd.merge(
        trips,
        tours_merged.drop(columns=["person_id", "household_id"]),
        left_on="tour_id",
        right_index=True,
        suffixes=["_trip", "_tour"],
        how="left",
    )

    # Add dataframes as local variables
    locals_d = {
        "persons": persons,
        "persons_merged": persons_merged,
        "households": households,
        "households_merged": households_merged,
        "trips": trips,
        "trips_merged": trips_merged,
        "tours": tours_merged,
        "tours_merged": tours_merged,
        "land_use": land_use,
    }

    skims = wrap_skims(network_los, trips_merged)

    # Annotate trips_merged
    expressions.annotate_preprocessors(
        trips_merged, locals_d, skims, model_settings, "summarize"
    )

    for table_name, df in locals_d.items():
        if table_name in model_settings:

            meta = model_settings[table_name]
            df = eval(table_name)

            if "AGGREGATE" in meta and meta["AGGREGATE"]:
                for agg in meta["AGGREGATE"]:
                    assert set(("column", "label", "map")) <= agg.keys()
                    df[agg["label"]] = (
                        df[agg["column"]].map(agg["map"]).fillna(df[agg["column"]])
                    )

            if "BIN" in meta and meta["BIN"]:
                for slicer in meta["BIN"]:
                    if slicer["type"] == "manual_breaks":
                        df[slicer["label"]] = manual_breaks(
                            df[slicer["column"]],
                            slicer["bin_breaks"],
                            slicer["bin_labels"],
                        )

                    elif slicer["type"] == "quantiles":
                        df[slicer["label"]] = quantiles(
                            df[slicer["column"]], slicer["bins"], slicer["label_format"]
                        )

                    elif slicer["type"] == "spaced_intervals":
                        df[slicer["label"]] = spaced_intervals(
                            df[slicer["column"]],
                            slicer["lower_bound"],
                            slicer["interval"],
                            slicer["label_format"],
                        )

                    elif slicer["type"] == "equal_intervals":
                        df[slicer["label"]] = equal_intervals(
                            df[slicer["column"]], slicer["bins"], slicer["label_format"]
                        )

    # Output pipeline tables for expression development
    if model_settings["EXPORT_PIPELINE_TABLES"] is True:
        pipeline_table_dir = os.path.join(output_location, "pipeline_tables")
        os.makedirs(config.output_file_path(pipeline_table_dir), exist_ok=True)
        for name, df in locals_d.items():
            df.to_csv(
                config.output_file_path(os.path.join(pipeline_table_dir, f"{name}.csv"))
            )

    # Add classification functions to locals
    locals_d.update(
        {
            "quantiles": quantiles,
            "spaced_intervals": spaced_intervals,
            "equal_intervals": equal_intervals,
            "manual_breaks": manual_breaks,
        }
    )

    for i, row in spec.iterrows():

        out_file = row["Output"]
        expr = row["Expression"]

        # Save temporary variables starting with underscores in locals_d
        if out_file.startswith("_"):

            logger.debug(f"Temp Variable: {expr} -> {out_file}")

            locals_d[out_file] = eval(expr, globals(), locals_d)
            continue

        logger.debug(f"Summary: {expr} -> {out_file}.csv")

        resultset = eval(expr, globals(), locals_d)
        resultset.to_csv(
            config.output_file_path(os.path.join(output_location, f"{out_file}.csv")),
            index=False,
        )



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate, tracing

from .util import estimation
from .util.tour_frequency import process_mandatory_tours

logger = logging.getLogger(__name__)


def add_null_results(trace_label, mandatory_tour_frequency_settings):
    logger.info("Skipping %s: add_null_results", trace_label)

    persons = inject.get_table("persons").to_frame()
    persons["mandatory_tour_frequency"] = ""

    tours = pd.DataFrame()
    tours["tour_category"] = None
    tours["tour_type"] = None
    tours["person_id"] = None
    tours.index.name = "tour_id"
    pipeline.replace_table("tours", tours)

    expressions.assign_columns(
        df=persons,
        model_settings=mandatory_tour_frequency_settings.get("annotate_persons"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_persons"),
    )

    pipeline.replace_table("persons", persons)


@inject.step()
def mandatory_tour_frequency(persons_merged, chunk_size, trace_hh_id):
    """
    This model predicts the frequency of making mandatory trips (see the
    alternatives above) - these trips include work and school in some combination.
    """
    trace_label = "mandatory_tour_frequency"
    model_settings_file_name = "mandatory_tour_frequency.yaml"

    model_settings = config.read_model_settings(model_settings_file_name)

    choosers = persons_merged.to_frame()
    # filter based on results of CDAP
    choosers = choosers[choosers.cdap_activity == "M"]
    logger.info("Running mandatory_tour_frequency with %d persons", len(choosers))

    # - if no mandatory tours
    if choosers.shape[0] == 0:
        add_null_results(trace_label, model_settings)
        return

    # - preprocessor
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:

        locals_dict = {}

        expressions.assign_columns(
            df=choosers,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

    estimator = estimation.manager.begin_estimation("mandatory_tour_frequency")

    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)
    constants = config.get_model_constants(model_settings)

    if estimator:
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_choosers(choosers)

    choices = simulate.simple_simulate(
        choosers=choosers,
        spec=model_spec,
        nest_spec=nest_spec,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="mandatory_tour_frequency",
        estimator=estimator,
    )

    # convert indexes to alternative names
    choices = pd.Series(model_spec.columns[choices.values], index=choices.index)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(
            choices, "persons", "mandatory_tour_frequency"
        )
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    # - create mandatory tours
    """
    This reprocesses the choice of index of the mandatory tour frequency
    alternatives into an actual dataframe of tours.  Ending format is
    the same as got non_mandatory_tours except trip types are "work" and "school"
    """
    alternatives = simulate.read_model_alts(
        "mandatory_tour_frequency_alternatives.csv", set_index="alt"
    )
    choosers["mandatory_tour_frequency"] = choices.reindex(choosers.index)

    mandatory_tours = process_mandatory_tours(
        persons=choosers, mandatory_tour_frequency_alts=alternatives
    )

    tours = pipeline.extend_table("tours", mandatory_tours)
    tracing.register_traceable_table("tours", mandatory_tours)
    pipeline.get_rn_generator().add_channel("tours", mandatory_tours)

    # - annotate persons
    persons = inject.get_table("persons").to_frame()

    # need to reindex as we only handled persons with cdap_activity == 'M'
    persons["mandatory_tour_frequency"] = (
        choices.reindex(persons.index).fillna("").astype(str)
    )

    expressions.assign_columns(
        df=persons,
        model_settings=model_settings.get("annotate_persons"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_persons"),
    )

    pipeline.replace_table("persons", persons)

    tracing.print_summary(
        "mandatory_tour_frequency", persons.mandatory_tour_frequency, value_counts=True
    )

    if trace_hh_id:
        tracing.trace_df(
            mandatory_tours,
            label="mandatory_tour_frequency.mandatory_tours",
            warn_if_empty=True,
        )

        tracing.trace_df(
            persons, label="mandatory_tour_frequency.persons", warn_if_empty=True
        )



# ActivitySim
# See full license in LICENSE.txt.
import logging
import warnings
from builtins import range

import numpy as np
import pandas as pd

from activitysim.core import (
    assign,
    chunk,
    config,
    expressions,
    inject,
    logit,
    los,
    pathbuilder_cache,
    simulate,
    tracing,
)
from activitysim.core.pathbuilder_cache import memo
from activitysim.core.util import reindex

logger = logging.getLogger(__name__)

TIMING = True
TRACE_CHUNK = True
ERR_CHECK = True
TRACE_COMPLEXITY = False  # diagnostic: log the omaz,dmaz pairs with the greatest number of virtual tap-tap paths

UNAVAILABLE = -999

# used as base file name for cached files and as shared buffer tag
CACHE_TAG = "tap_tap_utilities"


def compute_utilities(
    network_los,
    model_settings,
    choosers,
    model_constants,
    trace_label,
    trace=False,
    trace_column_names=None,
):
    """
    Compute utilities
    """
    trace_label = tracing.extend_trace_label(trace_label, "compute_utils")

    with chunk.chunk_log(trace_label):

        logger.debug(
            f"{trace_label} Running compute_utilities with {choosers.shape[0]} choosers"
        )

        locals_dict = {"np": np, "los": network_los}
        locals_dict.update(model_constants)

        # we don't grok coefficients, but allow them to use constants in spec alt columns
        spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
        for c in spec.columns:
            if c != simulate.SPEC_LABEL_NAME:
                spec[c] = spec[c].map(lambda s: model_constants.get(s, s)).astype(float)

        # - run preprocessor to annotate choosers
        preprocessor_settings = model_settings.get("PREPROCESSOR")
        if preprocessor_settings:

            # don't want to alter caller's dataframe
            choosers = choosers.copy()

            expressions.assign_columns(
                df=choosers,
                model_settings=preprocessor_settings,
                locals_dict=locals_dict,
                trace_label=trace_label,
            )

        utilities = simulate.eval_utilities(
            spec,
            choosers,
            locals_d=locals_dict,
            trace_all_rows=trace,
            trace_label=trace_label,
            trace_column_names=trace_column_names,
        )

    return utilities


class TransitVirtualPathBuilder(object):
    """
    Transit virtual path builder for three zone systems
    """

    def __init__(self, network_los):

        self.network_los = network_los

        self.uid_calculator = pathbuilder_cache.TapTapUidCalculator(network_los)

        # note: pathbuilder_cache is lightweight until opened
        self.tap_cache = pathbuilder_cache.TVPBCache(
            self.network_los, self.uid_calculator, CACHE_TAG
        )

        assert (
            network_los.zone_system == los.THREE_ZONE
        ), f"TransitVirtualPathBuilder: network_los zone_system not THREE_ZONE"

    def trace_df(self, df, trace_label, extension):
        assert len(df) > 0
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, extension),
            slicer="NONE",
            transpose=False,
        )

    def trace_maz_tap(self, maz_od_df, access_mode, egress_mode):
        def maz_tap_stats(mode, name):
            maz_tap_df = self.network_los.maz_to_tap_dfs[mode].reset_index()
            logger.debug(f"TVPB access_maz_tap {maz_tap_df.shape}")
            MAZ_count = len(maz_tap_df.MAZ.unique())
            TAP_count = len(maz_tap_df.TAP.unique())
            MAZ_PER_TAP = MAZ_count / TAP_count
            logger.debug(
                f"TVPB maz_tap_stats {name} {mode} MAZ {MAZ_count} TAP {TAP_count} ratio {MAZ_PER_TAP}"
            )

        logger.debug(f"TVPB maz_od_df {maz_od_df.shape}")

        maz_tap_stats(access_mode, "access")
        maz_tap_stats(egress_mode, "egress")

    def units_for_recipe(self, recipe):
        units = self.network_los.setting(f"TVPB_SETTINGS.{recipe}.units")
        assert units in [
            "utility",
            "time",
        ], f"unrecognized units: {units} for {recipe}. Expected either 'time' or 'utility'."
        return units

    def compute_maz_tap_utilities(
        self, recipe, maz_od_df, chooser_attributes, leg, mode, trace_label, trace
    ):

        trace_label = tracing.extend_trace_label(trace_label, f"maz_tap_utils.{leg}")

        with chunk.chunk_log(trace_label):
            maz_tap_settings = self.network_los.setting(
                f"TVPB_SETTINGS.{recipe}.maz_tap_settings.{mode}"
            )
            chooser_columns = maz_tap_settings["CHOOSER_COLUMNS"]
            attribute_columns = (
                list(chooser_attributes.columns)
                if chooser_attributes is not None
                else []
            )
            model_constants = self.network_los.setting(
                f"TVPB_SETTINGS.{recipe}.CONSTANTS"
            )

            if leg == "access":
                maz_col = "omaz"
                tap_col = "btap"
            else:
                maz_col = "dmaz"
                tap_col = "atap"

            # maz_to_tap access/egress utilities
            # deduped utilities_df - one row per chooser for each boarding tap (btap) accessible from omaz
            utilities_df = self.network_los.maz_to_tap_dfs[mode]

            utilities_df = (
                utilities_df[chooser_columns]
                .reset_index(drop=False)
                .rename(columns={"MAZ": maz_col, "TAP": tap_col})
            )
            utilities_df = pd.merge(
                maz_od_df[["idx", maz_col]].drop_duplicates(),
                utilities_df,
                on=maz_col,
                how="inner",
            )

            if len(utilities_df) == 0:
                trace = False
            # add any supplemental chooser attributes (e.g. demographic_segment, tod)
            for c in attribute_columns:
                utilities_df[c] = reindex(chooser_attributes[c], utilities_df["idx"])

            chunk.log_df(trace_label, "utilities_df", utilities_df)

            if self.units_for_recipe(recipe) == "utility":

                utilities_df[leg] = compute_utilities(
                    self.network_los,
                    maz_tap_settings,
                    utilities_df,
                    model_constants=model_constants,
                    trace_label=trace_label,
                    trace=trace,
                    trace_column_names=["idx", maz_col, tap_col] if trace else None,
                )

                chunk.log_df(trace_label, "utilities_df", utilities_df)  # annotated

            else:

                assignment_spec = assign.read_assignment_spec(
                    file_name=config.config_file_path(maz_tap_settings["SPEC"])
                )

                results, _, _ = assign.assign_variables(
                    assignment_spec, utilities_df, model_constants
                )
                assert len(results.columns == 1)
                utilities_df[leg] = results

            chunk.log_df(trace_label, "utilities_df", utilities_df)

            if trace:
                self.trace_df(utilities_df, trace_label, "utilities_df")

            # drop utility computation columns ('tod', 'demographic_segment' and maz_to_tap_df time/distance columns)
            utilities_df.drop(columns=attribute_columns + chooser_columns, inplace=True)

        return utilities_df

    def all_transit_paths(
        self, access_df, egress_df, chooser_attributes, trace_label, trace
    ):

        trace_label = tracing.extend_trace_label(trace_label, "all_transit_paths")

        # deduped transit_df has one row per chooser for each boarding (btap) and alighting (atap) pair
        transit_df = pd.merge(
            access_df[["idx", "btap"]], egress_df[["idx", "atap"]], on="idx"
        ).drop_duplicates()

        # don't want transit trips that start and stop in same tap
        transit_df = transit_df[transit_df.atap != transit_df.btap]

        for c in list(chooser_attributes.columns):
            transit_df[c] = reindex(chooser_attributes[c], transit_df["idx"])

        transit_df = transit_df.reset_index(drop=True)

        if trace:
            self.trace_df(transit_df, trace_label, "all_transit_df")

        return transit_df

    def compute_tap_tap_utilities(
        self,
        recipe,
        access_df,
        egress_df,
        chooser_attributes,
        path_info,
        trace_label,
        trace,
    ):
        """
        create transit_df and compute utilities for all atap-btap pairs between omaz in access and dmaz in egress_df
        compute the utilities using the tap_tap utility expressions file specified in tap_tap_settings

        transit_df contains all possible access omaz/btap to egress dmaz/atap transit path pairs for each chooser

        trace should be True as we don't encourage/support dynamic utility computation except when tracing
        (precompute being fairly fast)

        Parameters
        ----------
        recipe: str
           'recipe' key in network_los.yaml TVPB_SETTINGS e.g. tour_mode_choice
        access_df: pandas.DataFrame
            dataframe with 'idx' and 'omaz' columns
        egress_df: pandas.DataFrame
            dataframe with 'idx' and 'dmaz' columns
        chooser_attributes: dict
        path_info
        trace_label: str
        trace: boolean

        Returns
        -------
        transit_df: pandas.dataframe
        """

        assert trace

        trace_label = tracing.extend_trace_label(trace_label, "compute_tap_tap_utils")

        with chunk.chunk_log(trace_label):

            model_constants = self.network_los.setting(
                f"TVPB_SETTINGS.{recipe}.CONSTANTS"
            )
            tap_tap_settings = self.network_los.setting(
                f"TVPB_SETTINGS.{recipe}.tap_tap_settings"
            )

            with memo("#TVPB CACHE compute_tap_tap_utilities all_transit_paths"):
                transit_df = self.all_transit_paths(
                    access_df, egress_df, chooser_attributes, trace_label, trace
                )
                # note: transit_df index is arbitrary
            chunk.log_df(trace_label, "transit_df", transit_df)

            # FIXME some expressions may want to know access mode -
            locals_dict = path_info.copy()
            locals_dict.update(model_constants)

            # columns needed for compute_utilities
            chooser_columns = ["btap", "atap"] + list(chooser_attributes.columns)

            # deduplicate transit_df to unique_transit_df
            with memo("#TVPB compute_tap_tap_utilities deduplicate transit_df"):

                attribute_segments = self.network_los.setting(
                    "TVPB_SETTINGS.tour_mode_choice.tap_tap_settings.attribute_segments"
                )
                scalar_attributes = {
                    k: locals_dict[k]
                    for k in attribute_segments.keys()
                    if k not in transit_df
                }

                transit_df["uid"] = self.uid_calculator.get_unique_ids(
                    transit_df, scalar_attributes
                )

                unique_transit_df = transit_df.loc[
                    ~transit_df.uid.duplicated(), chooser_columns + ["uid"]
                ]
                logger.debug(
                    f"#TVPB CACHE deduped transit_df from {len(transit_df)} to {len(unique_transit_df)}"
                )

                unique_transit_df.set_index("uid", inplace=True)
                chunk.log_df(trace_label, "unique_transit_df", unique_transit_df)

                transit_df = transit_df[
                    ["idx", "btap", "atap", "uid"]
                ]  # don't need chooser columns
                chunk.log_df(trace_label, "transit_df", transit_df)

            logger.debug(
                f"#TVPB CACHE compute_tap_tap_utilities dedupe transit_df "
                f"from {len(transit_df)} to {len(unique_transit_df)} rows"
            )

            num_unique_transit_rows = len(unique_transit_df)  # errcheck
            logger.debug(
                f"#TVPB CACHE compute_tap_tap_utilities compute_utilities for {len(unique_transit_df)} rows"
            )

            with memo("#TVPB compute_tap_tap_utilities compute_utilities"):
                unique_utilities_df = compute_utilities(
                    self.network_los,
                    tap_tap_settings,
                    choosers=unique_transit_df,
                    model_constants=locals_dict,
                    trace_label=trace_label,
                    trace=trace,
                    trace_column_names=chooser_columns if trace else None,
                )
                chunk.log_df(trace_label, "unique_utilities_df", unique_utilities_df)
                chunk.log_df(
                    trace_label, "unique_transit_df", unique_transit_df
                )  # annotated

                if trace:
                    # combine unique_transit_df with unique_utilities_df for legibility
                    omnibus_df = pd.merge(
                        unique_transit_df,
                        unique_utilities_df,
                        left_index=True,
                        right_index=True,
                        how="left",
                    )
                    self.trace_df(omnibus_df, trace_label, "unique_utilities_df")
                    chunk.log_df(trace_label, "omnibus_df", omnibus_df)
                    del omnibus_df
                    chunk.log_df(trace_label, "omnibus_df", None)

            assert num_unique_transit_rows == len(unique_utilities_df)  # errcheck

            # redupe unique_transit_df back into transit_df
            with memo("#TVPB compute_tap_tap_utilities redupe transit_df"):

                # idx = transit_df.index
                transit_df = pd.merge(
                    transit_df, unique_utilities_df, left_on="uid", right_index=True
                )
                del transit_df["uid"]
                # transit_df.index = idx
                # note: left merge on columns does not preserve index,
                # but transit_df index is arbitrary so no need to restore

                chunk.log_df(trace_label, "transit_df", transit_df)

            for c in unique_utilities_df:
                assert ERR_CHECK and not transit_df[c].isnull().any()

            if len(unique_transit_df) > 0:
                # if all rows were cached, then unique_utilities_df is just a ref to cache
                del unique_utilities_df
                chunk.log_df(trace_label, "unique_utilities_df", None)

            chunk.log_df(trace_label, "transit_df", None)

            if trace:
                self.trace_df(transit_df, trace_label, "transit_df")

        return transit_df

    def lookup_tap_tap_utilities(
        self,
        recipe,
        maz_od_df,
        access_df,
        egress_df,
        chooser_attributes,
        path_info,
        trace_label,
    ):
        """
        create transit_df and compute utilities for all atap-btap pairs between omaz in access and dmaz in egress_df
        look up the utilities in the precomputed tap_cache data (which is indexed by uid_calculator unique_ids)
        (unique_id can used as a zero-based index into the data array)

        transit_df contains all possible access omaz/btap to egress dmaz/atap transit path pairs for each chooser

        Parameters
        ----------
        recipe
        maz_od_df
        access_df
        egress_df
        chooser_attributes
        path_info
        trace_label

        Returns
        -------

        """

        trace_label = tracing.extend_trace_label(trace_label, "lookup_tap_tap_utils")

        with chunk.chunk_log(trace_label):

            with memo("#TVPB CACHE lookup_tap_tap_utilities all_transit_paths"):
                transit_df = self.all_transit_paths(
                    access_df, egress_df, chooser_attributes, trace_label, trace=False
                )
                # note: transit_df index is arbitrary
                chunk.log_df(trace_label, "transit_df", transit_df)

            if TRACE_COMPLEXITY:
                # diagnostic: log the omaz,dmaz pairs with the greatest number of virtual tap-tap paths
                num_paths = transit_df.groupby(["idx"]).size().to_frame("n")
                num_paths = pd.merge(
                    maz_od_df, num_paths, left_on="idx", right_index=True
                )
                num_paths = num_paths[["omaz", "dmaz", "n"]].drop_duplicates(
                    subset=["omaz", "dmaz"]
                )
                num_paths = num_paths.sort_values("n", ascending=False).reset_index(
                    drop=True
                )
                logger.debug(f"num_paths\n{num_paths.head(10)}")

            # FIXME some expressions may want to know access mode -
            locals_dict = path_info.copy()

            # add uid column to transit_df
            with memo("#TVPB lookup_tap_tap_utilities assign uid"):
                attribute_segments = self.network_los.setting(
                    "TVPB_SETTINGS.tour_mode_choice.tap_tap_settings.attribute_segments"
                )
                scalar_attributes = {
                    k: locals_dict[k]
                    for k in attribute_segments.keys()
                    if k not in transit_df
                }

                transit_df.index = self.uid_calculator.get_unique_ids(
                    transit_df, scalar_attributes
                )
                transit_df = transit_df[
                    ["idx", "btap", "atap"]
                ]  # just needed chooser_columns for uid calculation
                chunk.log_df(trace_label, "transit_df add uid index", transit_df)

            with memo("#TVPB lookup_tap_tap_utilities reindex transit_df"):
                utilities = self.tap_cache.data
                i = 0
                for column_name in self.uid_calculator.set_names:
                    transit_df[column_name] = utilities[transit_df.index.values, i]
                    i += 1

            for c in self.uid_calculator.set_names:
                assert ERR_CHECK and not transit_df[c].isnull().any()

            chunk.log_df(trace_label, "transit_df", None)

        return transit_df

    def compute_tap_tap_time(
        self,
        recipe,
        access_df,
        egress_df,
        chooser_attributes,
        path_info,
        trace_label,
        trace,
    ):

        trace_label = tracing.extend_trace_label(trace_label, "compute_tap_tap_time")

        with chunk.chunk_log(trace_label):

            model_constants = self.network_los.setting(
                f"TVPB_SETTINGS.{recipe}.CONSTANTS"
            )
            tap_tap_settings = self.network_los.setting(
                f"TVPB_SETTINGS.{recipe}.tap_tap_settings"
            )

            with memo("#TVPB CACHE compute_tap_tap_utilities all_transit_paths"):
                transit_df = self.all_transit_paths(
                    access_df, egress_df, chooser_attributes, trace_label, trace
                )
                # note: transit_df index is arbitrary
                chunk.log_df(trace_label, "transit_df", transit_df)

            # some expressions may want to know access mode -
            locals_dict = path_info.copy()
            locals_dict["los"] = self.network_los
            locals_dict.update(model_constants)

            assignment_spec = assign.read_assignment_spec(
                file_name=config.config_file_path(tap_tap_settings["SPEC"])
            )

            DEDUPE = True
            if DEDUPE:

                # assign uid for reduping
                max_atap = transit_df.atap.max() + 1
                transit_df["uid"] = transit_df.btap * max_atap + transit_df.atap

                # dedupe
                chooser_attribute_columns = list(chooser_attributes.columns)
                unique_transit_df = transit_df.loc[
                    ~transit_df.uid.duplicated(),
                    ["btap", "atap", "uid"] + chooser_attribute_columns,
                ]
                unique_transit_df.set_index("uid", inplace=True)
                chunk.log_df(trace_label, "unique_transit_df", unique_transit_df)

                logger.debug(
                    f"#TVPB CACHE deduped transit_df from {len(transit_df)} to {len(unique_transit_df)}"
                )

                # assign_variables
                results, _, _ = assign.assign_variables(
                    assignment_spec, unique_transit_df, locals_dict
                )
                assert len(results.columns == 1)
                unique_transit_df["transit"] = results

                # redupe results back into transit_df
                with memo("#TVPB compute_tap_tap_time redupe transit_df"):
                    transit_df["transit"] = reindex(
                        unique_transit_df.transit, transit_df.uid
                    )

                del transit_df["uid"]
                del unique_transit_df
                chunk.log_df(trace_label, "transit_df", transit_df)
                chunk.log_df(trace_label, "unique_transit_df", None)

            else:
                results, _, _ = assign.assign_variables(
                    assignment_spec, transit_df, locals_dict
                )
                assert len(results.columns == 1)
                transit_df["transit"] = results

            # filter out unavailable btap_atap pairs
            logger.debug(
                f"{(transit_df['transit'] <= 0).sum()} unavailable tap_tap pairs out of {len(transit_df)}"
            )
            transit_df = transit_df[transit_df.transit > 0]

            transit_df.drop(columns=chooser_attributes.columns, inplace=True)

            chunk.log_df(trace_label, "transit_df", None)

            if trace:
                self.trace_df(transit_df, trace_label, "transit_df")

        return transit_df

    def compute_tap_tap(
        self,
        recipe,
        maz_od_df,
        access_df,
        egress_df,
        chooser_attributes,
        path_info,
        trace_label,
        trace,
    ):

        if self.units_for_recipe(recipe) == "utility":

            if not self.tap_cache.is_open:
                with memo("#TVPB compute_tap_tap tap_cache.open"):
                    self.tap_cache.open()

            if trace:
                result = self.compute_tap_tap_utilities(
                    recipe,
                    access_df,
                    egress_df,
                    chooser_attributes,
                    path_info,
                    trace_label,
                    trace,
                )
            else:
                result = self.lookup_tap_tap_utilities(
                    recipe,
                    maz_od_df,
                    access_df,
                    egress_df,
                    chooser_attributes,
                    path_info,
                    trace_label,
                )
            return result
        else:
            assert self.units_for_recipe(recipe) == "time"

            with memo("#TVPB compute_tap_tap_time"):
                result = self.compute_tap_tap_time(
                    recipe,
                    access_df,
                    egress_df,
                    chooser_attributes,
                    path_info,
                    trace_label,
                    trace,
                )
        return result

    def best_paths(
        self,
        recipe,
        path_type,
        maz_od_df,
        access_df,
        egress_df,
        transit_df,
        trace_label,
        trace=False,
    ):

        trace_label = tracing.extend_trace_label(trace_label, "best_paths")

        with chunk.chunk_log(trace_label):

            path_settings = self.network_los.setting(
                f"TVPB_SETTINGS.{recipe}.path_types.{path_type}"
            )
            max_paths_per_tap_set = path_settings.get("max_paths_per_tap_set", 1)
            max_paths_across_tap_sets = path_settings.get(
                "max_paths_across_tap_sets", 1
            )

            units = self.units_for_recipe(recipe)
            smaller_is_better = units in ["time"]

            maz_od_df["seq"] = maz_od_df.index
            # maz_od_df has one row per chooser
            # inner join to add rows for each access, egress, and transit segment combination
            path_df = (
                maz_od_df.merge(access_df, on=["idx", "omaz"], how="inner")
                .merge(egress_df, on=["idx", "dmaz"], how="inner")
                .merge(transit_df, on=["idx", "atap", "btap"], how="inner")
            )

            chunk.log_df(trace_label, "path_df", path_df)

            # transit sets are the transit_df non-join columns
            transit_sets = [
                c for c in transit_df.columns if c not in ["idx", "atap", "btap"]
            ]

            if trace:
                # be nice and show both tap_tap set utility and total_set = access + set + egress
                for c in transit_sets:
                    path_df[f"total_{c}"] = (
                        path_df[c] + path_df["access"] + path_df["egress"]
                    )
                self.trace_df(path_df, trace_label, "best_paths.full")
                for c in transit_sets:
                    del path_df[f"total_{c}"]

            for c in transit_sets:
                path_df[c] = path_df[c] + path_df["access"] + path_df["egress"]
            path_df.drop(columns=["access", "egress"], inplace=True)

            # choose best paths by tap set
            best_paths_list = []
            for c in transit_sets:
                keep = path_df.index.isin(
                    path_df[["seq", c]]
                    .sort_values(by=c, ascending=smaller_is_better)
                    .groupby(["seq"])
                    .head(max_paths_per_tap_set)
                    .index
                )

                best_paths_for_set = path_df[keep]
                best_paths_for_set["path_set"] = c  # remember the path set
                best_paths_for_set[units] = path_df[keep][c]
                best_paths_for_set.drop(columns=transit_sets, inplace=True)
                best_paths_list.append(best_paths_for_set)

            path_df = pd.concat(best_paths_list).sort_values(
                by=["seq", units], ascending=[True, smaller_is_better]
            )

            # choose best paths overall by seq
            path_df = path_df.sort_values(
                by=["seq", units], ascending=[True, smaller_is_better]
            )
            path_df = path_df[
                path_df.index.isin(
                    path_df.groupby(["seq"]).head(max_paths_across_tap_sets).index
                )
            ]

            if trace:
                self.trace_df(path_df, trace_label, "best_paths")

        return path_df

    def build_virtual_path(
        self,
        recipe,
        path_type,
        orig,
        dest,
        tod,
        demographic_segment,
        want_choices,
        trace_label,
        filter_targets=None,
        trace=False,
        override_choices=None,
    ):

        trace_label = tracing.extend_trace_label(trace_label, "build_virtual_path")

        # Tracing is implemented as a seperate, second call that operates ONLY on filter_targets
        assert not (trace and filter_targets is None)
        if filter_targets is not None:
            assert filter_targets.any()

            # slice orig and dest
            orig = orig[filter_targets]
            dest = dest[filter_targets]
            assert len(orig) > 0
            assert len(dest) > 0

            # slice tod and demographic_segment if not scalar
            if not isinstance(tod, str):
                tod = tod[filter_targets]
            if demographic_segment is not None:
                demographic_segment = demographic_segment[filter_targets]
                assert len(demographic_segment) > 0

            # slice choices
            # (requires actual choices from the previous call lest rands change on second call)
            assert want_choices == (override_choices is not None)
            if want_choices:
                override_choices = override_choices[filter_targets]

        units = self.units_for_recipe(recipe)
        assert (
            units == "utility" or not want_choices
        ), "'want_choices' only supported supported if units is utility"

        access_mode = self.network_los.setting(
            f"TVPB_SETTINGS.{recipe}.path_types.{path_type}.access"
        )
        egress_mode = self.network_los.setting(
            f"TVPB_SETTINGS.{recipe}.path_types.{path_type}.egress"
        )
        path_types_settings = self.network_los.setting(
            f"TVPB_SETTINGS.{recipe}.path_types.{path_type}"
        )
        attributes_as_columns = self.network_los.setting(
            f"TVPB_SETTINGS.{recipe}.tap_tap_settings.attributes_as_columns", []
        )

        path_info = {
            "path_type": path_type,
            "access_mode": access_mode,
            "egress_mode": egress_mode,
        }

        # maz od pairs requested
        with memo("#TVPB build_virtual_path maz_od_df"):
            maz_od_df = pd.DataFrame(
                {
                    "idx": orig.index.values,
                    "omaz": orig.values,
                    "dmaz": dest.values,
                    "seq": range(len(orig)),
                }
            )
            chunk.log_df(trace_label, "maz_od_df", maz_od_df)
            self.trace_maz_tap(maz_od_df, access_mode, egress_mode)

        # for location choice, there will be multiple alt dest rows per chooser and duplicate orig.index values
        # but tod and demographic_segment should be the same for all chooser rows (unique orig index values)
        # knowing this allows us to eliminate redundant computations (e.g. utilities of maz_tap pairs)
        duplicated = orig.index.duplicated(keep="first")
        chooser_attributes = pd.DataFrame(index=orig.index[~duplicated])
        if not isinstance(tod, str):
            chooser_attributes["tod"] = tod.loc[~duplicated]
        elif "tod" in attributes_as_columns:
            chooser_attributes["tod"] = tod
        else:
            path_info["tod"] = tod
        if demographic_segment is not None:
            chooser_attributes["demographic_segment"] = demographic_segment.loc[
                ~duplicated
            ]

        with memo("#TVPB build_virtual_path access_df"):
            access_df = self.compute_maz_tap_utilities(
                recipe,
                maz_od_df,
                chooser_attributes,
                leg="access",
                mode=access_mode,
                trace_label=trace_label,
                trace=trace,
            )
        chunk.log_df(trace_label, "access_df", access_df)

        with memo("#TVPB build_virtual_path egress_df"):
            egress_df = self.compute_maz_tap_utilities(
                recipe,
                maz_od_df,
                chooser_attributes,
                leg="egress",
                mode=egress_mode,
                trace_label=trace_label,
                trace=trace,
            )
        chunk.log_df(trace_label, "egress_df", egress_df)

        # L200 will drop all rows if all trips are intra-tap.
        if np.array_equal(access_df["btap"].values, egress_df["atap"].values):
            trace = False

        # path_info for use by expressions (e.g. penalty for drive access if no parking at access tap)
        with memo("#TVPB build_virtual_path compute_tap_tap"):
            if len(access_df) * len(egress_df) == 0:
                trace = False
            transit_df = self.compute_tap_tap(
                recipe,
                maz_od_df,
                access_df,
                egress_df,
                chooser_attributes,
                path_info=path_info,
                trace_label=trace_label,
                trace=trace,
            )
        chunk.log_df(trace_label, "transit_df", transit_df)

        # Cannot trace if df is empty. Prob happened at L200
        if len(transit_df) == 0:
            want_choices = False

        with memo("#TVPB build_virtual_path best_paths"):
            path_df = self.best_paths(
                recipe,
                path_type,
                maz_od_df,
                access_df,
                egress_df,
                transit_df,
                trace_label,
                trace,
            )
        chunk.log_df(trace_label, "path_df", path_df)

        # now that we have created path_df, we are done with the dataframes for the separate legs
        del access_df
        chunk.log_df(trace_label, "access_df", None)
        del egress_df
        chunk.log_df(trace_label, "egress_df", None)
        del transit_df
        chunk.log_df(trace_label, "transit_df", None)

        if units == "utility":

            # logsums
            with memo("#TVPB build_virtual_path logsums"):
                # one row per seq with utilities in columns
                # path_num 0-based to aligh with logit.make_choices 0-based choice indexes
                path_df["path_num"] = path_df.groupby("seq").cumcount()
                chunk.log_df(trace_label, "path_df", path_df)

                utilities_df = (
                    path_df[["seq", "path_num", units]]
                    .set_index(["seq", "path_num"])
                    .unstack()
                )
                utilities_df.columns = (
                    utilities_df.columns.droplevel()
                )  # for legibility

                # add rows missing because no access or egress availability
                utilities_df = pd.concat(
                    [pd.DataFrame(index=maz_od_df.seq), utilities_df], axis=1
                )
                utilities_df = utilities_df.fillna(
                    UNAVAILABLE
                )  # set utilities for missing paths to UNAVAILABLE

                chunk.log_df(trace_label, "utilities_df", utilities_df)

                with warnings.catch_warnings(record=True) as w:
                    # Cause all warnings to always be triggered.
                    # most likely "divide by zero encountered in log" caused by all transit sets non-viable
                    warnings.simplefilter("always")

                    paths_nest_nesting_coefficient = path_types_settings.get(
                        "paths_nest_nesting_coefficient", 1
                    )
                    exp_utilities = np.exp(
                        utilities_df.values / paths_nest_nesting_coefficient
                    )
                    logsums = np.maximum(
                        np.log(np.nansum(exp_utilities, axis=1)), UNAVAILABLE
                    )

                    if len(w) > 0:
                        for wrn in w:
                            logger.warning(
                                f"{trace_label} - {type(wrn).__name__} ({wrn.message})"
                            )

                        DUMP = False
                        if DUMP:
                            zero_utilities_df = utilities_df[
                                np.nansum(np.exp(utilities_df.values), axis=1) == 0
                            ]
                            zero_utilities_df.to_csv(
                                config.output_file_path("warning_utilities_df.csv"),
                                index=True,
                            )

            if want_choices:

                # orig index to identify appropriate random number channel to use making choices
                utilities_df.index = orig.index

                with memo("#TVPB build_virtual_path make_choices"):

                    probs = logit.utils_to_probs(
                        utilities_df, allow_zero_probs=True, trace_label=trace_label
                    )
                    chunk.log_df(trace_label, "probs", probs)

                    if trace:
                        choices = override_choices

                        utilities_df["choices"] = choices
                        self.trace_df(utilities_df, trace_label, "utilities_df")

                        probs["choices"] = choices
                        self.trace_df(probs, trace_label, "probs")
                    else:

                        choices, rands = logit.make_choices(
                            probs, allow_bad_probs=True, trace_label=trace_label
                        )

                        chunk.log_df(trace_label, "rands", rands)
                        del rands
                        chunk.log_df(trace_label, "rands", None)

                    del probs
                    chunk.log_df(trace_label, "probs", None)

                # we need to get path_set, btap, atap from path_df row with same seq and path_num
                # drop seq join column, but keep path_num of choice to override_choices when tracing
                columns_to_cache = ["btap", "atap", "path_set", "path_num"]
                logsum_df = (
                    pd.merge(
                        pd.DataFrame(
                            {"seq": range(len(orig)), "path_num": choices.values}
                        ),
                        path_df[["seq"] + columns_to_cache],
                        on=["seq", "path_num"],
                        how="left",
                    )
                    .drop(columns=["seq"])
                    .set_index(orig.index)
                )

                logsum_df["logsum"] = logsums

            else:

                assert len(logsums) == len(orig)
                logsum_df = pd.DataFrame({"logsum": logsums}, index=orig.index)

            chunk.log_df(trace_label, "logsum_df", logsum_df)

            del utilities_df
            chunk.log_df(trace_label, "utilities_df", None)

            if trace:
                self.trace_df(logsum_df, trace_label, "logsum_df")

            chunk.log_df(trace_label, "logsum_df", logsum_df)
            results = logsum_df

        else:
            assert units == "time"

            # return a series
            results = pd.Series(path_df[units].values, index=path_df["idx"])

            # zero-fill rows for O-D pairs where no best path exists because there was no tap-tap transit availability
            results = reindex(results, maz_od_df.idx).fillna(0.0)

            chunk.log_df(trace_label, "results", results)

        assert len(results) == len(orig)

        del path_df
        chunk.log_df(trace_label, "path_df", None)

        # diagnostic
        # maz_od_df['DIST'] = self.network_los.get_default_skim_dict().get('DIST').get(maz_od_df.omaz, maz_od_df.dmaz)
        # maz_od_df[units] = results.logsum if units == 'utility' else results.values
        # print(f"maz_od_df\n{maz_od_df}")

        return results

    def get_tvpb_logsum(
        self,
        path_type,
        orig,
        dest,
        tod,
        demographic_segment,
        want_choices,
        recipe="tour_mode_choice",
        trace_label=None,
    ):

        # assume they have given us a more specific name (since there may be more than one active wrapper)
        trace_label = trace_label or "get_tvpb_logsum"
        trace_label = tracing.extend_trace_label(trace_label, path_type)

        with chunk.chunk_log(trace_label):

            logsum_df = self.build_virtual_path(
                recipe,
                path_type,
                orig,
                dest,
                tod,
                demographic_segment,
                want_choices=want_choices,
                trace_label=trace_label,
            )

            trace_hh_id = inject.get_injectable("trace_hh_id", None)
            if (all(logsum_df["logsum"] == UNAVAILABLE)) or (len(logsum_df) == 0):
                trace_hh_id = False

            if trace_hh_id:
                filter_targets = tracing.trace_targets(orig)
                # choices from preceding run (because random numbers)
                override_choices = logsum_df["path_num"] if want_choices else None
                if filter_targets.any():
                    self.build_virtual_path(
                        recipe,
                        path_type,
                        orig,
                        dest,
                        tod,
                        demographic_segment,
                        want_choices=want_choices,
                        override_choices=override_choices,
                        trace_label=trace_label,
                        filter_targets=filter_targets,
                        trace=True,
                    )

        return logsum_df

    def get_tvpb_best_transit_time(self, orig, dest, tod):

        # FIXME lots of pathological knowledge here as we are only called by accessibility directly from expressions

        trace_label = tracing.extend_trace_label("accessibility.tvpb_best_time", tod)
        recipe = "accessibility"
        path_type = "WTW"

        with chunk.chunk_log(trace_label):
            result = self.build_virtual_path(
                recipe,
                path_type,
                orig,
                dest,
                tod,
                demographic_segment=None,
                want_choices=False,
                trace_label=trace_label,
            )

            trace_od = inject.get_injectable("trace_od", None)
            if trace_od:
                filter_targets = (orig == trace_od[0]) & (dest == trace_od[1])
                if filter_targets.any():
                    self.build_virtual_path(
                        recipe,
                        path_type,
                        orig,
                        dest,
                        tod,
                        demographic_segment=None,
                        want_choices=False,
                        trace_label=trace_label,
                        filter_targets=filter_targets,
                        trace=True,
                    )

        return result

    def wrap_logsum(
        self,
        orig_key,
        dest_key,
        tod_key,
        segment_key,
        recipe="tour_mode_choice",
        cache_choices=False,
        trace_label=None,
        tag=None,
    ):

        return TransitVirtualPathLogsumWrapper(
            self,
            orig_key,
            dest_key,
            tod_key,
            segment_key,
            recipe,
            cache_choices,
            trace_label,
            tag,
        )


class TransitVirtualPathLogsumWrapper(object):
    """
    Transit virtual path builder logsum wrapper for three zone systems
    """

    def __init__(
        self,
        pathbuilder,
        orig_key,
        dest_key,
        tod_key,
        segment_key,
        recipe,
        cache_choices,
        trace_label,
        tag,
    ):

        self.tvpb = pathbuilder
        assert hasattr(pathbuilder, "get_tvpb_logsum")

        self.orig_key = orig_key
        self.dest_key = dest_key
        self.tod_key = tod_key
        self.segment_key = segment_key
        self.recipe = recipe
        self.df = None

        self.cache_choices = cache_choices
        self.cache = {} if cache_choices else None

        self.base_trace_label = (
            tracing.extend_trace_label(trace_label, tag) or f"tvpb_logsum.{tag}"
        )
        self.trace_label = self.base_trace_label
        self.tag = tag

        assert isinstance(orig_key, str)
        assert isinstance(dest_key, str)
        assert isinstance(tod_key, str)
        assert isinstance(segment_key, str)

    def set_df(self, df):
        """
        Set the dataframe

        Parameters
        ----------
        df : DataFrame
            The dataframe which contains the origin and destination ids

        Returns
        -------
        self (to facilitiate chaining)
        """
        self.df = df
        return self

    def extend_trace_label(self, extension=None):
        if extension:
            self.trace_label = tracing.extend_trace_label(
                self.base_trace_label, extension
            )
        else:
            self.trace_label = self.base_trace_label

    def __getitem__(self, path_type):
        """
        Get an available skim object

        Parameters
        ----------
        key : hashable
             The key (identifier) for this skim object

        Returns
        -------
        skim: Skim
             The skim object
        """

        assert self.df is not None, "Call set_df first"
        assert (
            self.orig_key in self.df
        ), f"TransitVirtualPathLogsumWrapper: orig_key '{self.orig_key}' not in df"
        assert (
            self.dest_key in self.df
        ), f"TransitVirtualPathLogsumWrapper: dest_key '{self.dest_key}' not in df"
        assert (
            self.tod_key in self.df
        ), f"TransitVirtualPathLogsumWrapper: tod_key '{self.tod_key}' not in df"
        assert (
            self.segment_key in self.df
        ), f"TransitVirtualPathLogsumWrapper: segment_key '{self.segment_key}' not in df"

        orig = self.df[self.orig_key].astype("int")
        dest = self.df[self.dest_key].astype("int")
        tod = self.df[self.tod_key]
        segment = self.df[self.segment_key]

        logsum_df = self.tvpb.get_tvpb_logsum(
            path_type,
            orig,
            dest,
            tod,
            segment,
            want_choices=self.cache_choices,
            recipe=self.recipe,
            trace_label=self.trace_label,
        )

        if (self.cache_choices) and (not all(logsum_df["logsum"] == UNAVAILABLE)):

            # not tested on duplicate index because not currently needed
            # caching strategy does not require unique indexes but care would need to be taken to maintain alignment
            assert not orig.index.duplicated().any()

            # we only need to cache taps and path_set
            choices_df = logsum_df[["atap", "btap", "path_set"]]

            if path_type in self.cache:
                assert (
                    len(self.cache.get(path_type).index.intersection(logsum_df.index))
                    == 0
                )
                choices_df = pd.concat([self.cache.get(path_type), choices_df])

            self.cache[path_type] = choices_df

        return logsum_df.logsum



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import (
    config,
    expressions,
    inject,
    los,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.pathbuilder import TransitVirtualPathBuilder
from activitysim.core.util import assign_in_place

from .util import estimation
from .util.mode import run_tour_mode_choice_simulate

logger = logging.getLogger(__name__)


@inject.step()
def atwork_subtour_mode_choice(
    tours, persons_merged, network_los, chunk_size, trace_hh_id
):
    """
    At-work subtour mode choice simulate
    """

    trace_label = "atwork_subtour_mode_choice"

    model_settings_file_name = "tour_mode_choice.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    logsum_column_name = model_settings.get("MODE_CHOICE_LOGSUM_COLUMN_NAME")
    mode_column_name = "tour_mode"

    tours = tours.to_frame()
    subtours = tours[tours.tour_category == "atwork"]

    # - if no atwork subtours
    if subtours.shape[0] == 0:
        tracing.no_results(trace_label)
        return

    subtours_merged = pd.merge(
        subtours,
        persons_merged.to_frame(),
        left_on="person_id",
        right_index=True,
        how="left",
    )

    logger.info("Running %s with %d subtours" % (trace_label, subtours_merged.shape[0]))

    tracing.print_summary(
        "%s tour_type" % trace_label, subtours_merged.tour_type, value_counts=True
    )

    constants = {}
    constants.update(config.get_model_constants(model_settings))

    skim_dict = network_los.get_default_skim_dict()

    # setup skim keys
    orig_col_name = "workplace_zone_id"
    dest_col_name = "destination"
    out_time_col_name = "start"
    in_time_col_name = "end"
    odt_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="out_period"
    )
    dot_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="in_period"
    )
    odr_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=orig_col_name, dest_key=dest_col_name, dim3_key="in_period"
    )
    dor_skim_stack_wrapper = skim_dict.wrap_3d(
        orig_key=dest_col_name, dest_key=orig_col_name, dim3_key="out_period"
    )
    od_skim_stack_wrapper = skim_dict.wrap(orig_col_name, dest_col_name)

    skims = {
        "odt_skims": odt_skim_stack_wrapper,
        "dot_skims": dot_skim_stack_wrapper,
        "odr_skims": odr_skim_stack_wrapper,
        "dor_skims": dor_skim_stack_wrapper,
        "od_skims": od_skim_stack_wrapper,
        "orig_col_name": orig_col_name,
        "dest_col_name": dest_col_name,
        "out_time_col_name": out_time_col_name,
        "in_time_col_name": in_time_col_name,
    }

    if network_los.zone_system == los.THREE_ZONE:
        # fixme - is this a lightweight object?
        tvpb = network_los.tvpb

        tvpb_logsum_odt = tvpb.wrap_logsum(
            orig_key=orig_col_name,
            dest_key=dest_col_name,
            tod_key="out_period",
            segment_key="demographic_segment",
            cache_choices=True,
            trace_label=trace_label,
            tag="tvpb_logsum_odt",
        )
        tvpb_logsum_dot = tvpb.wrap_logsum(
            orig_key=dest_col_name,
            dest_key=orig_col_name,
            tod_key="in_period",
            segment_key="demographic_segment",
            cache_choices=True,
            trace_label=trace_label,
            tag="tvpb_logsum_dot",
        )

        skims.update(
            {"tvpb_logsum_odt": tvpb_logsum_odt, "tvpb_logsum_dot": tvpb_logsum_dot}
        )

        # TVPB constants can appear in expressions
        constants.update(
            network_los.setting("TVPB_SETTINGS.tour_mode_choice.CONSTANTS")
        )

    estimator = estimation.manager.begin_estimation("atwork_subtour_mode_choice")
    if estimator:
        estimator.write_coefficients(model_settings=model_settings)
        estimator.write_coefficients_template(model_settings=model_settings)
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        # FIXME run_tour_mode_choice_simulate writes choosers post-annotation

    choices_df = run_tour_mode_choice_simulate(
        subtours_merged,
        tour_purpose="atwork",
        model_settings=model_settings,
        mode_column_name=mode_column_name,
        logsum_column_name=logsum_column_name,
        network_los=network_los,
        skims=skims,
        constants=constants,
        estimator=estimator,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="tour_mode_choice",
    )

    # add cached tvpb_logsum tap choices for modes specified in tvpb_mode_path_types
    if network_los.zone_system == los.THREE_ZONE:

        tvpb_mode_path_types = model_settings.get("tvpb_mode_path_types")
        for mode, path_types in tvpb_mode_path_types.items():

            for direction, skim in zip(
                ["od", "do"], [tvpb_logsum_odt, tvpb_logsum_dot]
            ):

                path_type = path_types[direction]
                skim_cache = skim.cache[path_type]

                print(f"mode {mode} direction {direction} path_type {path_type}")

                for c in skim_cache:

                    dest_col = f"{direction}_{c}"

                    if dest_col not in choices_df:
                        choices_df[dest_col] = (
                            np.nan
                            if pd.api.types.is_numeric_dtype(skim_cache[c])
                            else ""
                        )

                    choices_df[dest_col].where(
                        choices_df.tour_mode != mode, skim_cache[c], inplace=True
                    )

    if estimator:
        estimator.write_choices(choices_df[mode_column_name])
        choices_df[mode_column_name] = estimator.get_survey_values(
            choices_df[mode_column_name], "tours", mode_column_name
        )
        estimator.write_override_choices(choices_df[mode_column_name])
        estimator.end_estimation()

    tracing.print_summary(
        "%s choices" % trace_label, choices_df[mode_column_name], value_counts=True
    )

    assign_in_place(tours, choices_df)
    pipeline.replace_table("tours", tours)

    # - annotate tours table
    if model_settings.get("annotate_tours"):
        tours = inject.get_table("tours").to_frame()
        expressions.assign_columns(
            df=tours,
            model_settings=model_settings.get("annotate_tours"),
            trace_label=tracing.extend_trace_label(trace_label, "annotate_tours"),
        )
        pipeline.replace_table("tours", tours)

    if trace_hh_id:
        tracing.trace_df(
            tours[tours.tour_category == "atwork"],
            label=tracing.extend_trace_label(trace_label, mode_column_name),
            slicer="tour_id",
            index_label="tour_id",
        )



# ActivitySim
# See full license in LICENSE.txt.
from . import (
    accessibility,
    atwork_subtour_destination,
    atwork_subtour_frequency,
    atwork_subtour_mode_choice,
    atwork_subtour_scheduling,
    auto_ownership,
    cdap,
    free_parking,
    initialize,
    initialize_los,
    initialize_tours,
    joint_tour_composition,
    joint_tour_destination,
    joint_tour_frequency,
    joint_tour_participation,
    joint_tour_scheduling,
    location_choice,
    mandatory_scheduling,
    mandatory_tour_frequency,
    non_mandatory_destination,
    non_mandatory_scheduling,
    non_mandatory_tour_frequency,
    parking_location_choice,
    stop_frequency,
    summarize,
    tour_mode_choice,
    tour_od_choice,
    tour_scheduling_probabilistic,
    trip_departure_choice,
    trip_destination,
    trip_matrices,
    trip_mode_choice,
    trip_purpose,
    trip_purpose_and_destination,
    trip_scheduling,
    trip_scheduling_choice,
    vehicle_allocation,
    vehicle_type_choice,
)



import multiprocessing

import numpy as np

from activitysim.benchmarking.componentwise import (
    template_component_timings_mp,
    template_setup_cache,
)

PRETTY_NAME = "MTC1_MP4"
EXAMPLE_NAME = "prototype_mtc_full"
NUM_PROCESSORS = int(np.clip(multiprocessing.cpu_count() - 2, 2, 4))
CONFIGS_DIRS = ("configs_mp", "configs")
DYNAMIC_CONFIG_DIR = "bench_configs_mp"
DATA_DIR = "data"
OUTPUT_DIR = "output_mp"
COMPONENT_NAMES = [
    "school_location",
    "workplace_location",
    "auto_ownership_simulate",
    "free_parking",
    "cdap_simulate",
    "mandatory_tour_frequency",
    "mandatory_tour_scheduling",
    "joint_tour_frequency",
    "joint_tour_composition",
    "joint_tour_participation",
    "joint_tour_destination",
    "joint_tour_scheduling",
    "non_mandatory_tour_frequency",
    "non_mandatory_tour_destination",
    "non_mandatory_tour_scheduling",
    "tour_mode_choice_simulate",
    "atwork_subtour_frequency",
    "atwork_subtour_destination",
    "atwork_subtour_scheduling",
    "atwork_subtour_mode_choice",
    "stop_frequency",
    "trip_purpose",
    "trip_destination",
    "trip_purpose_and_destination",
    "trip_scheduling",
    "trip_mode_choice",
]
BENCHMARK_SETTINGS = {
    # TODO: This multiprocess benchmarking is minimally functional,
    # but has a bad habit of crashing due to memory allocation errors on
    # all but the tiniest of examples. It would be great to fix the MP
    # benchmarks so they use chunking, automatically configure for available
    # RAM, and run a training-production cycle to get useful timing results.
    "households_sample_size": 400,
    "num_processes": NUM_PROCESSORS,
}
SKIM_CACHE = False
TIMEOUT = 36000.0  # ten hours
VERSION = "1"


def setup_cache():
    template_setup_cache(
        EXAMPLE_NAME,
        COMPONENT_NAMES,
        BENCHMARK_SETTINGS,
        dict(
            read_skim_cache=SKIM_CACHE,
            write_skim_cache=SKIM_CACHE,
        ),
        CONFIGS_DIRS,
        DATA_DIR,
        OUTPUT_DIR,
        config_overload_dir=DYNAMIC_CONFIG_DIR,
    )


template_component_timings_mp(
    globals(),
    COMPONENT_NAMES,
    EXAMPLE_NAME,
    OUTPUT_DIR,
    PRETTY_NAME,
    VERSION,
)



# ActivitySim
# See full license in LICENSE.txt.
from . import (
    accessibility,
    households,
    landuse,
    persons,
    shadow_pricing,
    size_terms,
    skims,
    table_dict,
    time_windows,
    tours,
    trips,
    vehicles,
)



import os
from pathlib import Path
from typing import Collection

import numpy as np
import pandas as pd
import yaml
from larch import DataFrames, Model, P, X
from larch.util import Dict

from .general import (
    apply_coefficients,
    clean_values,
    construct_nesting_tree,
    explicit_value_parameters,
    linear_utility_from_spec,
    remove_apostrophes,
)
from .simple_simulate import construct_availability, simple_simulate_data


def mode_choice_model(
    name,
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
    override_filenames=None,
):
    if override_filenames is None:
        override_filenames = {}
    edb_directory = edb_directory.format(name=name)
    data = simple_simulate_data(
        name=name,
        edb_directory=edb_directory,
        **override_filenames,
    )
    coefficients = data.coefficients
    coef_template = data.coef_template
    spec = data.spec
    chooser_data = data.chooser_data
    settings = data.settings

    chooser_data = clean_values(
        chooser_data,
        alt_names_to_codes=data.alt_names_to_codes,
        choice_code="override_choice_code",
    )

    tree = construct_nesting_tree(data.alt_names, settings["NESTS"])

    purposes = list(coef_template.columns)
    if "atwork" in name:
        purposes = ["atwork"]
    elif "atwork" in purposes:
        purposes.remove("atwork")

    # Setup purpose specific models
    m = {purpose: Model(graph=tree, title=purpose) for purpose in purposes}
    for alt_code, alt_name in tree.elemental_names().items():
        # Read in base utility function for this alt_name
        u = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=alt_name,
            ignore_x=("#",),
        )
        for purpose in purposes:
            # Modify utility function based on template for purpose
            u_purp = sum(
                (P(coef_template[purpose].get(i.param, i.param)) * i.data * i.scale)
                for i in u
            )
            m[purpose].utility_co[alt_code] = u_purp

    for model in m.values():
        explicit_value_parameters(model)
    apply_coefficients(coefficients, m)

    avail = construct_availability(
        m[purposes[0]], chooser_data, data.alt_codes_to_names
    )

    d = DataFrames(
        co=chooser_data,
        av=avail,
        alt_codes=data.alt_codes,
        alt_names=data.alt_names,
    )

    if "atwork" not in name:
        for purpose, model in m.items():
            model.dataservice = d.selector_co(f"tour_type=='{purpose}'")
            model.choice_co_code = "override_choice_code"
    else:
        for purpose, model in m.items():
            model.dataservice = d
            model.choice_co_code = "override_choice_code"

    from larch.model.model_group import ModelGroup

    mg = ModelGroup(m.values())

    if return_data:
        return (
            mg,
            Dict(
                edb_directory=Path(edb_directory),
                chooser_data=chooser_data,
                avail=avail,
                coefficients=coefficients,
                coef_template=coef_template,
                spec=spec,
                settings=settings,
            ),
        )

    return mg


def tour_mode_choice_model(
    name="tour_mode_choice",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return mode_choice_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
    )


def trip_mode_choice_model(
    name="trip_mode_choice",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return mode_choice_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
    )


def atwork_subtour_mode_choice_model(
    name="atwork_subtour_mode_choice",
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    return mode_choice_model(
        name=name,
        edb_directory=edb_directory,
        return_data=return_data,
        override_filenames=dict(
            coefficients_file="tour_mode_choice_coefficients.csv",
        ),
    )



# ActivitySim
# See full license in LICENSE.txt.


import os

import pandas as pd
import pandas.testing as pdt
import pytest

from ..tour_frequency import process_non_mandatory_tours


def test_nmtf():

    persons = pd.DataFrame(
        {
            "non_mandatory_tour_frequency": [0, 3, 2, 1],
            "household_id": [1, 1, 2, 4],
            "home_zone_id": [100, 100, 200, 400],
        },
        index=[0, 1, 2, 3],
    )

    non_mandatory_tour_frequency_alts = pd.DataFrame(
        {"escort": [0, 0, 2, 0], "shopping": [1, 0, 0, 0], "othmaint": [0, 1, 0, 0]},
        index=[0, 1, 2, 3],
    )

    tour_counts = non_mandatory_tour_frequency_alts.loc[
        persons.non_mandatory_tour_frequency
    ]
    tour_counts.index = persons.index  # assign person ids to the index

    # - create the non_mandatory tours
    nmt = process_non_mandatory_tours(persons, tour_counts)

    idx = nmt.index

    pdt.assert_series_equal(
        nmt.person_id, pd.Series([0, 2, 2, 3], index=idx, name="person_id")
    )

    pdt.assert_series_equal(
        nmt.tour_type,
        pd.Series(
            ["shopping", "escort", "escort", "othmaint"], index=idx, name="tour_type"
        ),
    )



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate
from activitysim.core import timetable as tt
from activitysim.core import tracing
from activitysim.core.util import assign_in_place, reindex

from .util import estimation
from .util import vectorize_tour_scheduling as vts
from .util.tour_scheduling import run_tour_scheduling

logger = logging.getLogger(__name__)

DUMP = False


@inject.step()
def mandatory_tour_scheduling(tours, persons_merged, tdd_alts, chunk_size, trace_hh_id):
    """
    This model predicts the departure time and duration of each activity for mandatory tours
    """

    model_name = "mandatory_tour_scheduling"
    trace_label = model_name

    persons_merged = persons_merged.to_frame()

    tours = tours.to_frame()
    mandatory_tours = tours[tours.tour_category == "mandatory"]

    # - if no mandatory_tours
    if mandatory_tours.shape[0] == 0:
        tracing.no_results(model_name)
        return

    # - add tour segmentation column
    # mtctm1 segments mandatory_scheduling spec by tour_type
    # (i.e. there are different specs for work and school tour_types)
    # mtctm1 logsum coefficients are segmented by primary_purpose
    # (i.e. there are different logsum coefficients for work, school, univ primary_purposes
    # for simplicity managing these different segmentation schemes,
    # we conflate them by segmenting tour processing to align with primary_purpose
    tour_segment_col = "mandatory_tour_seg"
    assert tour_segment_col not in mandatory_tours
    is_university_tour = (mandatory_tours.tour_type == "school") & reindex(
        persons_merged.is_university, mandatory_tours.person_id
    )
    mandatory_tours[tour_segment_col] = mandatory_tours.tour_type.where(
        ~is_university_tour, "univ"
    )

    choices = run_tour_scheduling(
        model_name,
        mandatory_tours,
        persons_merged,
        tdd_alts,
        tour_segment_col,
        chunk_size,
        trace_hh_id,
    )

    assign_in_place(tours, choices)
    pipeline.replace_table("tours", tours)

    # updated df for tracing
    mandatory_tours = tours[tours.tour_category == "mandatory"]

    tracing.dump_df(
        DUMP,
        tt.tour_map(persons_merged, mandatory_tours, tdd_alts),
        trace_label,
        "tour_map",
    )

    if trace_hh_id:
        tracing.trace_df(
            mandatory_tours,
            label=trace_label,
            slicer="person_id",
            index_label="tour",
            columns=None,
            warn_if_empty=True,
        )
